import uuid

import pytest

from ffun.core.tests.helpers import TableSizeDelta, TableSizeNotChanged
from ffun.scores import domain, errors, operations


class TestCreateOrUpdateRule:
    @pytest.mark.asyncio
    async def test_create_new_rule(self, internal_user_id: uuid.UUID, three_tags_ids: tuple[int, int, int]) -> None:
        async with TableSizeDelta("s_rules", delta=1):
            created_rule = await operations.create_or_update_rule(internal_user_id, three_tags_ids, 13)

        rules = await domain.get_rules(internal_user_id)

        assert rules == [created_rule]

        assert created_rule.user_id == internal_user_id
        assert created_rule.tags == set(three_tags_ids)
        assert created_rule.score == 13

    @pytest.mark.asyncio
    async def test_tags_order_does_not_affect_creation(
        self, internal_user_id: uuid.UUID, three_tags_ids: tuple[int, int, int]
    ) -> None:
        async with TableSizeDelta("s_rules", delta=1):
            created_rule_1 = await operations.create_or_update_rule(internal_user_id, three_tags_ids, 13)
            created_rule_2 = await operations.create_or_update_rule(internal_user_id, reversed(three_tags_ids), 17)

        assert created_rule_1.tags == created_rule_2.tags

        rules = await domain.get_rules(internal_user_id)

        assert rules == [created_rule_2]

    @pytest.mark.asyncio
    async def test_update_scores_of_existed_rule(
        self, internal_user_id: uuid.UUID, three_tags_ids: tuple[int, int, int]
    ) -> None:
        await operations.create_or_update_rule(internal_user_id, three_tags_ids, 13)

        async with TableSizeNotChanged("s_rules"):
            updated_rule = await operations.create_or_update_rule(internal_user_id, three_tags_ids, 17)

        rules = await domain.get_rules(internal_user_id)

        assert rules == [updated_rule]

        assert updated_rule.user_id == internal_user_id
        assert updated_rule.tags == set(three_tags_ids)
        assert updated_rule.score == 17

    @pytest.mark.asyncio
    async def test_multiple_entities(
        self, internal_user_id: uuid.UUID, another_internal_user_id: uuid.UUID, three_tags_ids: tuple[int, int, int]
    ) -> None:
        await operations.create_or_update_rule(internal_user_id, three_tags_ids[:2], 3)
        await operations.create_or_update_rule(another_internal_user_id, three_tags_ids[:2], 5)
        await operations.create_or_update_rule(another_internal_user_id, three_tags_ids[1:], 7)
        await operations.create_or_update_rule(internal_user_id, three_tags_ids[:2], 11)

        rules = await domain.get_rules(internal_user_id)

        assert len(rules) == 1

        assert rules[0].user_id == internal_user_id
        assert rules[0].tags == set(three_tags_ids[:2])
        assert rules[0].score == 11

        rules = await domain.get_rules(another_internal_user_id)

        rules.sort(key=lambda r: r.score)

        assert len(rules) == 2

        assert rules[0].user_id == another_internal_user_id
        assert rules[0].tags == set(three_tags_ids[:2])
        assert rules[0].score == 5

        assert rules[1].user_id == another_internal_user_id
        assert rules[1].tags == set(three_tags_ids[1:])
        assert rules[1].score == 7


class TestDeleteRule:
    @pytest.mark.asyncio
    async def test_delete_rule(
        self, internal_user_id: uuid.UUID, another_internal_user_id: uuid.UUID, three_tags_ids: tuple[int, int, int]
    ) -> None:
        rule_to_delete = await operations.create_or_update_rule(internal_user_id, three_tags_ids, 13)
        rule_2 = await operations.create_or_update_rule(internal_user_id, three_tags_ids[:2], 17)
        rule_3 = await operations.create_or_update_rule(another_internal_user_id, three_tags_ids, 19)

        async with TableSizeDelta("s_rules", delta=-1):
            await operations.delete_rule(internal_user_id, rule_to_delete.id)

        rules = await domain.get_rules(internal_user_id)

        assert rules == [rule_2]

        rules = await domain.get_rules(another_internal_user_id)

        assert rules == [rule_3]

    @pytest.mark.asyncio
    async def test_delete_not_existed_rule(
        self, internal_user_id: uuid.UUID, three_tags_ids: tuple[int, int, int]
    ) -> None:
        async with TableSizeNotChanged("s_rules"):
            await operations.delete_rule(internal_user_id, uuid.uuid4())

    @pytest.mark.asyncio
    async def test_delete_for_wrong_user(
        self, internal_user_id: uuid.UUID, another_internal_user_id: uuid.UUID, three_tags_ids: tuple[int, int, int]
    ) -> None:
        rule_to_delete = await operations.create_or_update_rule(internal_user_id, three_tags_ids, 13)

        async with TableSizeNotChanged("s_rules"):
            await operations.delete_rule(another_internal_user_id, rule_to_delete.id)


class TestUpdateRule:
    @pytest.mark.asyncio
    async def test_update_rule(self, internal_user_id: uuid.UUID, three_tags_ids: tuple[int, int, int]) -> None:
        rule_to_update = await operations.create_or_update_rule(internal_user_id, three_tags_ids, 13)

        async with TableSizeNotChanged("s_rules"):
            updated_rule = await operations.update_rule(internal_user_id, rule_to_update.id, three_tags_ids[:2], 17)

        assert updated_rule.id == rule_to_update.id
        assert updated_rule.user_id == internal_user_id
        assert updated_rule.tags == set(three_tags_ids[:2])
        assert updated_rule.score == 17

        rules = await domain.get_rules(internal_user_id)

        assert rules == [updated_rule]

    @pytest.mark.asyncio
    async def test_update_not_existed_rule(
        self, internal_user_id: uuid.UUID, three_tags_ids: tuple[int, int, int]
    ) -> None:
        async with TableSizeNotChanged("s_rules"):
            with pytest.raises(errors.NoRuleFound):
                await operations.update_rule(internal_user_id, uuid.uuid4(), three_tags_ids[:2], 17)

    @pytest.mark.asyncio
    async def test_wrong_user(
        self, internal_user_id: uuid.UUID, another_internal_user_id: uuid.UUID, three_tags_ids: tuple[int, int, int]
    ) -> None:
        rule_to_update = await operations.create_or_update_rule(internal_user_id, three_tags_ids, 13)

        async with TableSizeNotChanged("s_rules"):
            with pytest.raises(errors.NoRuleFound):
                await operations.update_rule(another_internal_user_id, rule_to_update.id, three_tags_ids, 17)

        rules = await domain.get_rules(internal_user_id)

        assert rules == [rule_to_update]


# most of the logic of this function is validated in other tests
class TestGetRules:
    @pytest.mark.asyncio
    async def test_no_rules(self, internal_user_id: uuid.UUID) -> None:
        rules = await domain.get_rules(internal_user_id)

        assert rules == []



import uuid

import pytest

from ffun.core.postgresql import transaction
from ffun.core.tests.helpers import TableSizeDelta, TableSizeNotChanged
from ffun.library.entities import Entry
from ffun.library.tests import make as l_make
from ffun.markers.entities import Marker
from ffun.markers.operations import get_markers, remove_markers_for_entries, set_marker, tech_merge_markers
from ffun.users.tests import make as u_make


class TestMergeFeeds:
    @pytest.mark.asyncio
    async def test_nothing_to_merge(self, new_entry: Entry, another_new_entry: Entry) -> None:
        async with transaction() as trx:
            await tech_merge_markers(trx, new_entry.id, another_new_entry.id)

    @pytest.mark.asyncio
    async def test_move_markers(self, loaded_feed_id: uuid.UUID) -> None:
        entry_1, entry_2, entry_3 = await l_make.n_entries_list(loaded_feed_id, 3)

        user_a, user_b = await u_make.n_users(2)

        await set_marker(user_a, Marker.read, entry_1.id)
        await set_marker(user_b, Marker.read, entry_2.id)

        async with TableSizeNotChanged("m_markers"):
            async with transaction() as trx:
                await tech_merge_markers(trx, from_entry_id=entry_2.id, to_entry_id=entry_3.id)

        markers_a = await get_markers(user_a, [entry_1.id, entry_2.id, entry_3.id])
        markers_b = await get_markers(user_b, [entry_1.id, entry_2.id, entry_3.id])

        assert markers_a == {entry_1.id: {Marker.read}}
        assert markers_b == {entry_3.id: {Marker.read}}

    @pytest.mark.asyncio
    async def test_rewrite_links(self, loaded_feed_id: uuid.UUID) -> None:
        entry_1, entry_2, entry_3 = await l_make.n_entries_list(loaded_feed_id, 3)

        user_a, user_b = await u_make.n_users(2)

        await set_marker(user_a, Marker.read, entry_1.id)
        await set_marker(user_b, Marker.read, entry_2.id)
        await set_marker(user_b, Marker.read, entry_3.id)

        async with TableSizeDelta("m_markers", delta=-1):
            async with transaction() as trx:
                await tech_merge_markers(trx, from_entry_id=entry_2.id, to_entry_id=entry_3.id)

        markers_a = await get_markers(user_a, [entry_1.id, entry_2.id, entry_3.id])
        markers_b = await get_markers(user_b, [entry_1.id, entry_2.id, entry_3.id])

        assert markers_a == {entry_1.id: {Marker.read}}
        assert markers_b == {entry_3.id: {Marker.read}}

    @pytest.mark.asyncio
    async def test_complex_ops(self, loaded_feed_id: uuid.UUID) -> None:
        entry_1, entry_2, entry_3 = await l_make.n_entries_list(loaded_feed_id, 3)

        user_a, user_b, user_c = await u_make.n_users(3)

        await set_marker(user_a, Marker.read, entry_1.id)

        await set_marker(user_b, Marker.read, entry_2.id)
        await set_marker(user_b, Marker.read, entry_3.id)

        await set_marker(user_c, Marker.read, entry_2.id)

        async with TableSizeDelta("m_markers", delta=-1):
            async with transaction() as trx:
                await tech_merge_markers(trx, from_entry_id=entry_2.id, to_entry_id=entry_3.id)

        markers_a = await get_markers(user_a, [entry_1.id, entry_2.id, entry_3.id])
        markers_b = await get_markers(user_b, [entry_1.id, entry_2.id, entry_3.id])
        markers_c = await get_markers(user_c, [entry_1.id, entry_2.id, entry_3.id])

        assert markers_a == {entry_1.id: {Marker.read}}
        assert markers_b == {entry_3.id: {Marker.read}}
        assert markers_c == {entry_3.id: {Marker.read}}


class TestRemoveMarkersForEntries:
    @pytest.mark.asyncio
    async def test_nothing_to_remove(self, loaded_feed_id: uuid.UUID) -> None:
        entry_1, entry_2, entry_3 = await l_make.n_entries_list(loaded_feed_id, 3)

        async with TableSizeNotChanged("m_markers"):
            await remove_markers_for_entries([entry_1.id, entry_2.id, entry_3.id])

    @pytest.mark.asyncio
    async def test_remove(self, loaded_feed_id: uuid.UUID) -> None:
        entry_1, entry_2, entry_3 = await l_make.n_entries_list(loaded_feed_id, 3)

        user_a, user_b, user_c = await u_make.n_users(3)

        await set_marker(user_a, Marker.read, entry_1.id)

        await set_marker(user_b, Marker.read, entry_2.id)
        await set_marker(user_b, Marker.read, entry_3.id)

        await set_marker(user_c, Marker.read, entry_2.id)

        async with TableSizeDelta("m_markers", delta=-3):
            await remove_markers_for_entries([entry_1.id, entry_2.id])

        markers_a = await get_markers(user_a, [entry_1.id, entry_2.id, entry_3.id])
        markers_b = await get_markers(user_b, [entry_1.id, entry_2.id, entry_3.id])
        markers_c = await get_markers(user_c, [entry_1.id, entry_2.id, entry_3.id])

        assert markers_a == {}
        assert markers_b == {entry_3.id: {Marker.read}}
        assert markers_c == {}



import uuid
from typing import Any

from ffun.core import logging
from ffun.core.postgresql import ExecuteType, execute
from ffun.feeds_links.entities import FeedLink

logger = logging.get_module_logger()


def row_to_feed_link(row: dict[str, Any]) -> FeedLink:
    return FeedLink(user_id=row["user_id"], feed_id=row["feed_id"], created_at=row["created_at"])


# TODO: tests
async def add_link(user_id: uuid.UUID, feed_id: uuid.UUID) -> None:
    sql = """
        INSERT INTO fl_links (id, user_id, feed_id)
        VALUES (%(id)s, %(user_id)s, %(feed_id)s)
        ON CONFLICT (user_id, feed_id) DO NOTHING
    """

    await execute(sql, {"id": uuid.uuid4(), "user_id": user_id, "feed_id": feed_id})


# TODO: tests
async def remove_link(user_id: uuid.UUID, feed_id: uuid.UUID) -> None:
    sql = """
        DELETE FROM fl_links WHERE user_id = %(user_id)s AND feed_id = %(feed_id)s
    """

    await execute(sql, {"user_id": user_id, "feed_id": feed_id})


# TODO: tests
async def get_linked_feeds(user_id: uuid.UUID) -> list[FeedLink]:
    sql = """
        SELECT * FROM fl_links WHERE user_id = %(user_id)s
    """

    result = await execute(sql, {"user_id": user_id})

    return [row_to_feed_link(row) for row in result]


# TODO: tests
async def get_linked_users(feed_id: uuid.UUID) -> list[uuid.UUID]:
    sql = """
        SELECT user_id FROM fl_links WHERE feed_id = %(feed_id)s
    """

    result = await execute(sql, {"feed_id": feed_id})

    return [row["user_id"] for row in result]


# TODO: tests
async def has_linked_users(feed_id: uuid.UUID) -> bool:
    sql = """
        SELECT 1 FROM fl_links WHERE feed_id = %(feed_id)s LIMIT 1
    """

    result = await execute(sql, {"feed_id": feed_id})

    return bool(result)


async def tech_merge_feeds(execute: ExecuteType, from_feed_id: uuid.UUID, to_feed_id: uuid.UUID) -> None:
    sql = """
    DELETE FROM fl_links as fll
    WHERE feed_id = %(from_feed_id)s
          AND EXISTS (SELECT 1 FROM fl_links WHERE feed_id = %(to_feed_id)s AND user_id = fll.user_id)
    """

    await execute(sql, {"from_feed_id": from_feed_id, "to_feed_id": to_feed_id})

    sql = """
    UPDATE fl_links
    SET feed_id = %(to_feed_id)s
    WHERE feed_id = %(from_feed_id)s
    """

    await execute(sql, {"from_feed_id": from_feed_id, "to_feed_id": to_feed_id})



import asyncio
from typing import AsyncGenerator, Generator

import fastapi
import pytest
import pytest_asyncio

from ffun.application import application
from ffun.core import migrations
from ffun.feeds.tests.fixtures import *  # noqa
from ffun.feeds_collections.tests.fixtures import *  # noqa
from ffun.librarian.tests.fixtures import *  # noqa
from ffun.library.tests.fixtures import *  # noqa
from ffun.ontology.tests.fixtures import *  # noqa
from ffun.openai.tests.fixtures import *  # noqa
from ffun.users.tests.fixtures import *  # noqa


@pytest.fixture(scope="session", autouse=True)
def event_loop() -> Generator[asyncio.AbstractEventLoop, asyncio.AbstractEventLoop, None]:
    loop = asyncio.get_event_loop_policy().new_event_loop()
    yield loop
    loop.close()


@pytest_asyncio.fixture(scope="session", autouse=True)
async def app() -> AsyncGenerator[fastapi.FastAPI, None]:
    async with application.with_app() as app:
        yield app


@pytest_asyncio.fixture(scope="session", autouse=True)
async def prepare_db(
    app: AsyncGenerator[fastapi.FastAPI, None],
    event_loop: asyncio.AbstractEventLoop,
) -> AsyncGenerator[None, None]:
    # database migrations may be in an inconsistent state
    await migrations.rollback_all()

    await migrations.apply_all()
    yield
    await migrations.rollback_all()



from ffun.feeds_collections.collections import artificial_intelligence, gamedev
from ffun.feeds_collections.entities import Collection

# temporary solution to fill feeds for a new user
#
# - engilish feeds only (for now)
# - no podcasts
# - no news aggregators
# - https-only
# - only alive
# - lists are highly subjective :-)

# TODO: define explicit rules
# TODO: for each source add short "why added" description and show it in UI

predefines = {
    Collection.gamedev: gamedev.collection,
    Collection.artificial_intelligence: artificial_intelligence.collection,
}



import os

import pytest

from ffun.parsers.entities import FeedInfo
from ffun.parsers.feed import parse_feed

feeds_fixtures_directory = os.path.join(os.path.dirname(__file__), "feeds_fixtures")


_excluded_suffixes = [".expected.json", "~", "#"]


def _feeds_fixtures_names() -> list[str]:
    files = os.listdir(feeds_fixtures_directory)

    return [filename for filename in files if not any(filename.endswith(suffix) for suffix in _excluded_suffixes)]


class TestParseFeed:
    @pytest.mark.parametrize("raw_fixture_name", _feeds_fixtures_names())
    def test_on_row_fixtures(self, raw_fixture_name: str) -> None:
        raw_fixture_path = os.path.join(feeds_fixtures_directory, raw_fixture_name)
        expected_fixture_path = raw_fixture_path + ".expected.json"

        with open(raw_fixture_path, "r", encoding="utf-8") as raw_fixture_file:
            raw_fixture = raw_fixture_file.read()

        with open(expected_fixture_path, "r", encoding="utf-8") as expected_fixture_file:
            expected_fixture = expected_fixture_file.read()

        feed_info = parse_feed(raw_fixture, "https://example.com/feed/")

        assert feed_info == FeedInfo.model_validate_json(expected_fixture)



collection = {
    "https://neurohive.io/en/feed/",
    "http://feeds.feedburner.com/MachineLearningMastery",
    "https://www.fast.ai/index.xml",
    "http://ruder.io/rss/index.rss",
    "https://www.reddit.com/r/MachineLearning/.rss",
    "https://embracethered.com/blog/index.xml",
    "https://www.reddit.com/r/alife/.rss",
    "https://googleresearch.blogspot.com/atom.xml",
    "https://blog.einstein.ai/rss/",
    "https://www.louisbouchard.ai/rss/",
    "http://hunch.net/?feed=rss2",
}



import asyncio
import random
import uuid

import pytest

from ffun.core import utils
from ffun.core.tests.helpers import TableSizeDelta
from ffun.feeds import errors
from ffun.feeds.domain import get_feed, save_feeds
from ffun.feeds.entities import Feed, FeedError, FeedState
from ffun.feeds.operations import (
    get_feeds,
    get_next_feeds_to_load,
    mark_feed_as_failed,
    mark_feed_as_loaded,
    mark_feed_as_orphaned,
    save_feed,
    tech_remove_feed,
    update_feed_info,
)
from ffun.feeds.tests import make


class TestSaveFeed:
    @pytest.mark.asyncio
    async def test_new_feed(self, new_feed: Feed) -> None:
        feed_id = await save_feed(new_feed)
        assert feed_id == new_feed.id
        assert await get_feed(new_feed.id) == new_feed

    @pytest.mark.asyncio
    async def test_existed_feed(self, new_feed: Feed) -> None:
        original_feed_id = await save_feed(new_feed)

        cloned_feed = new_feed.replace(id=uuid.uuid4())

        saved_feed_id = await save_feed(cloned_feed)

        assert original_feed_id == saved_feed_id

        assert await get_feed(original_feed_id) == new_feed

        with pytest.raises(errors.NoFeedFound):
            await get_feed(cloned_feed.id)

    @pytest.mark.asyncio
    async def test_same_uid_different_urls(self, new_feed: Feed) -> None:
        url_part = uuid.uuid4().hex

        feed_1 = new_feed.replace(url=f"http://example.com/{url_part}")
        feed_2 = new_feed.replace(id=uuid.uuid4(), url=f"https://example.com/{url_part}")

        feed_1_id = await save_feed(feed_1)
        feed_2_id = await save_feed(feed_2)

        assert feed_1_id == feed_2_id

        assert await get_feed(feed_1_id) == feed_1

        with pytest.raises(errors.NoFeedFound):
            await get_feed(feed_2.id)


class TestGetNextFeedToLoad:
    @pytest.mark.asyncio
    async def test_find_new_feed(self, saved_feed_id: uuid.UUID) -> None:
        now = utils.now()

        found_feeds = await get_next_feeds_to_load(number=100500, loaded_before=now)

        feeds = {feed.id: feed for feed in found_feeds}

        found_feed = feeds[saved_feed_id]

        assert now < found_feed.load_attempted_at

        loaded_feed = await get_feed(saved_feed_id)

        assert loaded_feed.load_attempted_at == found_feed.load_attempted_at

    @pytest.mark.asyncio
    async def test_skip_choosen_feeds(self, saved_feed_id: uuid.UUID) -> None:
        now = utils.now()

        await get_next_feeds_to_load(number=100500, loaded_before=now)

        found_feeds = await get_next_feeds_to_load(number=100500, loaded_before=now)

        assert not found_feeds

    @pytest.mark.asyncio
    async def test_find_choosen_after_time_passes(self, saved_feed_id: uuid.UUID) -> None:
        await get_next_feeds_to_load(number=100500, loaded_before=utils.now())

        now = utils.now()

        found_feeds = await get_next_feeds_to_load(number=100500, loaded_before=now)

        feeds = {feed.id: feed for feed in found_feeds}

        found_feed = feeds[saved_feed_id]

        assert now < found_feed.load_attempted_at

        loaded_feed = await get_feed(saved_feed_id)

        assert loaded_feed.load_attempted_at == found_feed.load_attempted_at

    @pytest.mark.asyncio
    async def test_limit(self) -> None:
        now = utils.now()

        # mark existed feeds as chosen
        await get_next_feeds_to_load(number=100500, loaded_before=now)

        n = 10
        m = 3

        feed_ids = set(await save_feeds([make.fake_feed() for _ in range(n)]))

        while feed_ids:
            found_feeds = await get_next_feeds_to_load(number=m, loaded_before=now)

            assert len(found_feeds) == min(m, len(feed_ids))

            found_feed_ids = {feed.id for feed in found_feeds}

            assert found_feed_ids <= feed_ids

            feed_ids -= found_feed_ids

    @pytest.mark.asyncio
    async def test_concurent_run(self) -> None:
        now = utils.now()

        # mark existed feeds as chosen
        await get_next_feeds_to_load(number=100500, loaded_before=now)

        n = 10
        m = 3

        feed_ids = set(await save_feeds([make.fake_feed() for _ in range(n)]))

        tasks = [get_next_feeds_to_load(number=m, loaded_before=now) for _ in range(n // m + 1)]

        results = await asyncio.gather(*tasks)

        found_feed_ids = {feed.id for feeds in results for feed in feeds}

        assert found_feed_ids == feed_ids


class TestUpdateFeedInfo:
    @pytest.mark.asyncio
    async def test(self, saved_feed: Feed) -> None:
        new_title = make.fake_title()
        new_description = make.fake_description()

        await update_feed_info(feed_id=saved_feed.id, title=new_title, description=new_description)

        updated_feed = await get_feed(saved_feed.id)

        assert updated_feed.title == new_title
        assert updated_feed.description == new_description


class TestMarkFeedAsLoaded:
    @pytest.mark.asyncio
    async def test(self, saved_feed: Feed) -> None:
        now = utils.now()

        await mark_feed_as_loaded(feed_id=saved_feed.id)

        updated_feed = await get_feed(saved_feed.id)

        assert updated_feed.loaded_at is not None
        assert now < updated_feed.loaded_at
        assert updated_feed.last_error is None
        assert updated_feed.state == FeedState.loaded

    @pytest.mark.asyncio
    async def test_reset_error_state(self, saved_feed: Feed) -> None:
        await mark_feed_as_failed(feed_id=saved_feed.id, state=FeedState.damaged, error=random.choice(list(FeedError)))

        await mark_feed_as_loaded(feed_id=saved_feed.id)

        updated_feed = await get_feed(saved_feed.id)

        assert updated_feed.last_error is None


class TestMarkFeedAsFailed:
    @pytest.mark.asyncio
    async def test(self, saved_feed: Feed) -> None:
        error = random.choice(list(FeedError))

        await mark_feed_as_failed(feed_id=saved_feed.id, state=FeedState.damaged, error=error)

        updated_feed = await get_feed(saved_feed.id)

        assert updated_feed.state == FeedState.damaged
        assert updated_feed.last_error == error


class TestMarkFeedAsOrphaned:
    @pytest.mark.asyncio
    async def test(self, saved_feed: Feed) -> None:
        await mark_feed_as_orphaned(feed_id=saved_feed.id)

        assert saved_feed.loaded_at is None

        updated_feed = await get_feed(saved_feed.id)

        assert updated_feed.loaded_at is None
        assert updated_feed.last_error is None
        assert updated_feed.state == FeedState.orphaned

    @pytest.mark.asyncio
    async def test_reset_error_state(self, saved_feed: Feed) -> None:
        await mark_feed_as_failed(feed_id=saved_feed.id, state=FeedState.damaged, error=random.choice(list(FeedError)))

        await mark_feed_as_orphaned(feed_id=saved_feed.id)

        updated_feed = await get_feed(saved_feed.id)

        assert updated_feed.last_error is None


class TestGetFeeds:
    @pytest.mark.asyncio
    async def test(self) -> None:
        n = 3

        feed_ids = []

        for _ in range(n + 2):
            raw_feed = make.fake_feed()
            feed_id = await save_feed(raw_feed)
            feed_ids.append(feed_id)

        loaded_feeds = await get_feeds(feed_ids[1:-1])

        assert len(loaded_feeds) == n

        assert set(feed_ids[1:-1]) == {feed.id for feed in loaded_feeds}


class TestTechRemoveFeed:
    @pytest.mark.asyncio
    async def test(self, saved_feed: Feed) -> None:
        async with TableSizeDelta("f_feeds", delta=-1):
            await tech_remove_feed(saved_feed.id)

        with pytest.raises(errors.NoFeedFound):
            await get_feed(saved_feed.id)

    @pytest.mark.asyncio
    async def test_no_feed(self) -> None:
        await tech_remove_feed(uuid.uuid4())



import uuid
from typing import Any, Iterable

import psycopg

from ffun.core import logging
from ffun.core.postgresql import execute
from ffun.scores import errors
from ffun.scores.entities import Rule

logger = logging.get_module_logger()


def _normalize_tags(tags: Iterable[int]) -> list[int]:
    return list(sorted(tags))


def _key_from_tags(tags: Iterable[int]) -> str:
    return ",".join(map(str, tags))


def row_to_rule(row: dict[str, Any]) -> Rule:
    return Rule(
        id=row["id"],
        user_id=row["user_id"],
        tags=set(row["tags"]),
        score=row["score"],
        created_at=row["created_at"],
        updated_at=row["updated_at"],
    )


async def create_or_update_rule(user_id: uuid.UUID, tags: Iterable[int], score: int) -> Rule:
    tags = _normalize_tags(tags)

    key = _key_from_tags(tags)

    sql = """
        INSERT INTO s_rules (id, user_id, tags, key, score)
        VALUES (%(id)s, %(user_id)s, %(tags)s, %(key)s, %(score)s)
        RETURNING *
        """

    try:
        result = await execute(sql, {"id": uuid.uuid4(), "user_id": user_id, "tags": tags, "key": key, "score": score})
    except psycopg.errors.UniqueViolation:
        logger.info("rule_already_exists_change_score", key=key)

        sql = """
        UPDATE s_rules
        SET score = %(score)s, updated_at = NOW()
        WHERE user_id = %(user_id)s AND key = %(key)s
        RETURNING *
        """

        result = await execute(sql, {"user_id": user_id, "key": key, "score": score})

    return row_to_rule(result[0])


async def delete_rule(user_id: uuid.UUID, rule_id: uuid.UUID) -> None:
    sql = """
        DELETE FROM s_rules
        WHERE user_id = %(user_id)s AND id = %(rule_id)s
        """

    await execute(sql, {"user_id": user_id, "rule_id": rule_id})


async def update_rule(user_id: uuid.UUID, rule_id: uuid.UUID, tags: Iterable[int], score: int) -> Rule:
    tags = _normalize_tags(tags)
    key = _key_from_tags(tags)

    sql = """
    UPDATE s_rules
    SET tags = %(tags)s, key = %(key)s, score = %(score)s, updated_at = NOW()
    WHERE user_id = %(user_id)s AND id = %(rule_id)s
    returning *
    """

    result = await execute(sql, {"user_id": user_id, "rule_id": rule_id, "tags": tags, "key": key, "score": score})

    if not result:
        raise errors.NoRuleFound()

    return row_to_rule(result[0])


async def get_rules(user_id: uuid.UUID) -> list[Rule]:
    sql = """
    SELECT *
    FROM s_rules
    WHERE user_id = %(user_id)s
    """

    rows = await execute(sql, {"user_id": user_id})

    return [row_to_rule(row) for row in rows]



import datetime
import enum
import uuid

import pydantic


class KeyStatus(str, enum.Enum):
    works = "works"
    broken = "broken"
    quota = "quota"
    unknown = "unknown"


class APIKeyUsage(pydantic.BaseModel):
    user_id: uuid.UUID
    api_key: str
    used_tokens: int | None


class OpenAIAnswer(pydantic.BaseModel):
    content: str

    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


class UserKeyInfo(pydantic.BaseModel):
    user_id: uuid.UUID
    api_key: str | None
    max_tokens_in_month: int
    process_entries_not_older_than: datetime.timedelta
    tokens_used: int



import datetime
import uuid

import pytest

from ffun.core import utils
from ffun.core.postgresql import execute
from ffun.core.tests.helpers import TableSizeDelta, TableSizeNotChanged
from ffun.librarian import errors, operations
from ffun.librarian.entities import ProcessorPointer
from ffun.librarian.tests import helpers
from ffun.library.tests import make as l_make


def assert_is_new_pointer(pointer: ProcessorPointer, processor_id: int) -> None:
    assert pointer.processor_id == processor_id
    assert pointer.pointer_created_at == datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    assert pointer.pointer_entry_id == uuid.UUID("00000000-0000-0000-0000-000000000000")


class TestCreatePointer:
    @pytest.mark.asyncio
    async def test_new_pointer(self, fake_processor_id: int) -> None:
        await operations.delete_pointer(fake_processor_id)

        async with TableSizeDelta("ln_processor_pointers", delta=1):
            pointer = await operations.create_pointer(fake_processor_id)

        assert pointer is not None

        assert_is_new_pointer(pointer, fake_processor_id)

        loaded_pointer = await operations.get_or_create_pointer(fake_processor_id)

        assert loaded_pointer.processor_id == fake_processor_id

    @pytest.mark.asyncio
    async def test_existing_pointer(self, fake_processor_id: int) -> None:
        pointer = await operations.get_or_create_pointer(fake_processor_id)

        async with TableSizeNotChanged("ln_processor_pointers"):
            second_pointer = await operations.create_pointer(fake_processor_id)

        assert second_pointer is None

        loaded_pointer = await operations.get_or_create_pointer(fake_processor_id)

        assert pointer == loaded_pointer


# the base behavior of the get_or_create_pointer function is checked in other tests, that use it
class TestGetOrCreatePointer:
    @pytest.mark.asyncio
    async def test_create_if_not_exists(self, fake_processor_id: int) -> None:
        await operations.delete_pointer(fake_processor_id)

        async with TableSizeDelta("ln_processor_pointers", delta=1):
            pointer = await operations.get_or_create_pointer(fake_processor_id)

        assert_is_new_pointer(pointer, fake_processor_id)

        loaded_pointer = await operations.get_or_create_pointer(fake_processor_id)

        assert loaded_pointer.processor_id == fake_processor_id


class TestDeletePointer:
    @pytest.mark.asyncio
    async def test_existing_pointer(self, fake_processor_id: int) -> None:
        pointer = await operations.get_or_create_pointer(fake_processor_id)

        pointer.replace(pointer_entry_id=uuid.uuid4())

        await operations.save_pointer(execute, pointer)

        async with TableSizeDelta("ln_processor_pointers", delta=-1):
            await operations.delete_pointer(fake_processor_id)

        new_pointer = await operations.get_or_create_pointer(fake_processor_id)

        assert_is_new_pointer(new_pointer, fake_processor_id)

    @pytest.mark.asyncio
    async def test_no_pointer(self, fake_processor_id: int) -> None:
        await operations.delete_pointer(fake_processor_id)

        async with TableSizeNotChanged("ln_processor_pointers"):
            await operations.delete_pointer(fake_processor_id)


class TestSavePointer:
    @pytest.mark.asyncio
    async def test_no_pointer(self, fake_processor_id: int) -> None:
        await operations.delete_pointer(fake_processor_id)

        pointer = ProcessorPointer(
            processor_id=fake_processor_id,
            pointer_entry_id=uuid.uuid4(),
            pointer_created_at=datetime.datetime.now(),
        )

        async with TableSizeNotChanged("ln_processor_pointers"):
            with pytest.raises(errors.CanNotSaveUnexistingPointer):
                await operations.save_pointer(execute, pointer)

    @pytest.mark.asyncio
    async def test_existing_pointer(self, fake_processor_id: int) -> None:
        pointer = await operations.get_or_create_pointer(fake_processor_id)

        new_pointer = pointer.replace(pointer_created_at=utils.now(), pointer_entry_id=uuid.uuid4())

        async with TableSizeNotChanged("ln_processor_pointers"):
            await operations.save_pointer(execute, new_pointer)

        loaded_pointer = await operations.get_or_create_pointer(fake_processor_id)

        assert new_pointer == loaded_pointer


class TestPushEntriesToProcessorQueue:
    @pytest.mark.parametrize("entries_count", [0, 1, 3, 10])
    @pytest.mark.asyncio
    async def test_push_entry(self, loaded_feed_id: uuid.UUID, entries_count: int, fake_processor_id: int) -> None:
        await operations.clear_processor_queue(fake_processor_id)

        entries = await l_make.n_entries(loaded_feed_id, n=entries_count + 5)

        entries_to_push = list(entries)[:entries_count]

        async with TableSizeDelta("ln_processors_queue", delta=entries_count):
            await operations.push_entries_to_processor_queue(execute, fake_processor_id, entries_to_push)

        entries_in_queue = await operations.get_entries_to_process(fake_processor_id, limit=entries_count + 5)

        assert set(entries_to_push) == set(entries_in_queue)


class TestGetEntriesToProcess:
    @pytest.mark.asyncio
    async def test_get_entries(self, loaded_feed_id: uuid.UUID, fake_processor_id: int) -> None:
        await operations.clear_processor_queue(fake_processor_id)

        entries = await l_make.n_entries(loaded_feed_id, n=13)

        await operations.push_entries_to_processor_queue(execute, fake_processor_id, list(entries))

        received_entries = set()

        while True:
            new_entries_ids = await operations.get_entries_to_process(fake_processor_id, limit=3)
            received_entries.update(new_entries_ids)
            await operations.remove_entries_from_processor_queue(execute, fake_processor_id, new_entries_ids)

            if len(received_entries) == len(entries):
                assert len(new_entries_ids) == 13 % 3
                break

        assert set(entries) == received_entries

    @pytest.mark.asyncio
    async def test_no_entries(self, fake_processor_id: int) -> None:
        await operations.clear_processor_queue(fake_processor_id)

        entries = await operations.get_entries_to_process(fake_processor_id, limit=3)

        assert not entries

    @pytest.mark.asyncio
    async def test_idempotency(self, loaded_feed_id: uuid.UUID, fake_processor_id: int) -> None:
        await operations.clear_processor_queue(fake_processor_id)

        entries = await l_make.n_entries(loaded_feed_id, n=13)

        await operations.push_entries_to_processor_queue(execute, fake_processor_id, list(entries))

        entries_ids_1 = await operations.get_entries_to_process(fake_processor_id, limit=3)
        entries_ids_2 = await operations.get_entries_to_process(fake_processor_id, limit=3)

        assert entries_ids_1 == entries_ids_2


class TestCountEntriesInProcessorQueue:
    @pytest.mark.asyncio
    async def test_no_entries(self, fake_processor_id: int) -> None:
        await operations.clear_processor_queue(fake_processor_id)

        count = await operations.count_entries_in_processor_queue(fake_processor_id)

        assert count == 0

    @pytest.mark.asyncio
    async def test_count(
        self, loaded_feed_id: uuid.UUID, fake_processor_id: int, another_fake_processor_id: int
    ) -> None:
        old_count = await operations.count_entries_in_processor_queue(fake_processor_id)

        entries = await l_make.n_entries(loaded_feed_id, n=3)
        more_entries = await l_make.n_entries(loaded_feed_id, n=2)

        await operations.push_entries_to_processor_queue(execute, fake_processor_id, list(entries))
        await operations.push_entries_to_processor_queue(execute, another_fake_processor_id, list(more_entries))

        new_count = await operations.count_entries_in_processor_queue(fake_processor_id)

        assert new_count == old_count + 3


class TestRemoveEntriesFromProcessorQueue:
    @pytest.mark.asyncio
    async def test_remove_entries(self, loaded_feed_id: uuid.UUID, fake_processor_id: int) -> None:
        await operations.clear_processor_queue(fake_processor_id)

        entries = await l_make.n_entries(loaded_feed_id, n=13)

        await operations.push_entries_to_processor_queue(execute, fake_processor_id, list(entries))

        entries_to_remove = list(entries)[:5]

        async with TableSizeDelta("ln_processors_queue", delta=-5):
            await operations.remove_entries_from_processor_queue(execute, fake_processor_id, entries_to_remove)

        entries_in_queue = await operations.get_entries_to_process(fake_processor_id, limit=13)

        assert len(entries_in_queue) == 8

        assert set(entries_to_remove) & set(entries_in_queue) == set()

    @pytest.mark.asyncio
    async def test_remove_non_existing_entries(self, loaded_feed_id: uuid.UUID, fake_processor_id: int) -> None:
        await operations.clear_processor_queue(fake_processor_id)

        entries = await l_make.n_entries(loaded_feed_id, n=13)

        await operations.push_entries_to_processor_queue(execute, fake_processor_id, list(entries))

        entries_to_remove = [uuid.uuid4(), uuid.uuid4()]

        await operations.remove_entries_from_processor_queue(execute, fake_processor_id, entries_to_remove)

        entries_in_queue = await operations.get_entries_to_process(fake_processor_id, limit=13)

        assert set(entries) == set(entries_in_queue)


class TestClearProcessorQueue:
    # checked in other tests
    pass


class TestAddEntriesToFailedStorage:
    @pytest.mark.asyncio
    async def test_add_entries(self, loaded_feed_id: uuid.UUID, fake_processor_id: int) -> None:
        entries = await l_make.n_entries(loaded_feed_id, n=13)

        entries_to_add = list(entries)[:5]

        async with TableSizeDelta("ln_failed_entries", delta=5):
            await operations.add_entries_to_failed_storage(fake_processor_id, entries_to_add)

        failed_entries = await operations.get_failed_entries(execute, fake_processor_id, limit=100500)

        assert set(entries_to_add) <= set(failed_entries)


class TestGetFailedEntries:
    @pytest.mark.asyncio
    async def test_get_entries(self, loaded_feed_id: uuid.UUID, fake_processor_id: int) -> None:
        entries = await l_make.n_entries(loaded_feed_id, n=4)

        await operations.add_entries_to_failed_storage(fake_processor_id, list(entries))

        failed_entries = await operations.get_failed_entries(execute, fake_processor_id, limit=100500)

        assert set(entries) <= set(failed_entries)

    @pytest.mark.asyncio
    async def test_no_entries(self, fake_processor_id: int) -> None:
        while True:
            failed_entries = await operations.get_failed_entries(execute, fake_processor_id, limit=1000)
            await operations.remove_failed_entries(execute, fake_processor_id, failed_entries)

            if not failed_entries:
                break

        failed_entries = await operations.get_failed_entries(execute, fake_processor_id, limit=100500)

        assert not failed_entries

    @pytest.mark.asyncio
    async def test_idempotency(self, loaded_feed_id: uuid.UUID, fake_processor_id: int) -> None:
        entries = await l_make.n_entries(loaded_feed_id, n=13)

        await operations.add_entries_to_failed_storage(fake_processor_id, list(entries))

        failed_entries_1 = await operations.get_failed_entries(execute, fake_processor_id, limit=100500)
        failed_entries_2 = await operations.get_failed_entries(execute, fake_processor_id, limit=100500)

        assert failed_entries_1 == failed_entries_2


class TestRemoveFailedEntries:
    @pytest.mark.asyncio
    async def test_remove_entries(self, loaded_feed_id: uuid.UUID, fake_processor_id: int) -> None:
        entries = await l_make.n_entries(loaded_feed_id, n=13)

        await operations.add_entries_to_failed_storage(fake_processor_id, list(entries))

        entries_to_remove = list(entries)[:5]

        async with TableSizeDelta("ln_failed_entries", delta=-5):
            await operations.remove_failed_entries(execute, fake_processor_id, entries_to_remove)

        failed_entries = await operations.get_failed_entries(execute, fake_processor_id, limit=100500)

        assert set(entries_to_remove) & set(failed_entries) == set()

    @pytest.mark.asyncio
    async def test_remove_non_existing_entries(self, loaded_feed_id: uuid.UUID, fake_processor_id: int) -> None:
        entries = await l_make.n_entries(loaded_feed_id, n=5)

        await operations.add_entries_to_failed_storage(fake_processor_id, list(entries))

        entries_to_remove = [uuid.uuid4(), uuid.uuid4()]

        async with TableSizeNotChanged("ln_failed_entries"):
            await operations.remove_failed_entries(execute, fake_processor_id, entries_to_remove)


class TestCountFailedEntries:
    @pytest.mark.asyncio
    async def test_no_entries(self, fake_processor_id: int, another_fake_processor_id: int) -> None:
        await helpers.clean_failed_storage([fake_processor_id, another_fake_processor_id])

        counts = await operations.count_failed_entries()

        assert fake_processor_id not in counts
        assert another_fake_processor_id not in counts

    @pytest.mark.asyncio
    async def test_some_entries(self, fake_processor_id: int, another_fake_processor_id: int) -> None:
        await helpers.clean_failed_storage([fake_processor_id, another_fake_processor_id])

        entries = await l_make.n_entries(uuid.uuid4(), n=3)
        entries_list = list(entries)

        await operations.add_entries_to_failed_storage(fake_processor_id, [entries_list[0], entries_list[1]])
        await operations.add_entries_to_failed_storage(
            another_fake_processor_id, [entry_id for entry_id in entries_list]
        )

        counts = await operations.count_failed_entries()

        assert counts[fake_processor_id] == 2
        assert counts[another_fake_processor_id] == 3



import uuid
from typing import Iterable

from ffun.core import logging, postgresql
from ffun.feeds import domain as f_domain
from ffun.feeds_links import domain as fl_domain
from ffun.library import domain as l_domain
from ffun.markers import domain as m_domain
from ffun.meta.settings import settings
from ffun.ontology import domain as o_domain

logger = logging.get_module_logger()


async def merge_feeds(feed_1_id: uuid.UUID, feed_2_id: uuid.UUID) -> None:
    """Merge feed_2 into feed_1, remove feed_2."""
    log = logger.bind(function="merge_feeds")

    log.info("start", feed_1_id=feed_1_id, feed_2_id=feed_2_id)

    all_entries = await l_domain.get_entries_by_filter(feeds_ids=[feed_1_id, feed_2_id], limit=postgresql.MAX_INTEGER)

    feed_1_entries = {entry.external_id: entry for entry in all_entries if entry.feed_id == feed_1_id}
    feed_2_entries = {entry.external_id: entry for entry in all_entries if entry.feed_id == feed_2_id}

    log.info("entries", feed_1=len(feed_1_entries), feed_2=len(feed_2_entries))

    # move missing entries from feed_2 to feed_1
    for external_id, entry in list(feed_2_entries.items()):
        log.info("process_entry", external_id=external_id, entry_id=entry.id)

        if external_id in feed_1_entries:
            log.info("entry_will_be_merged", entry_id=entry.id)
            continue

        log.info("move_entry", entry_id=entry.id)
        await l_domain.tech_move_entry(entry.id, feed_1_id)

        del feed_2_entries[external_id]

    log.info("merge_entries")

    # move tags from entries in feed_2 to entries in feed_1
    for entry_2 in feed_2_entries.values():
        entry_1 = feed_1_entries[entry_2.external_id]

        log.info("copy_relations", from_entry_id=entry_2.id, to_entry_id=entry_1.id)
        await o_domain.tech_copy_relations(entry_from_id=entry_2.id, entry_to_id=entry_1.id)

        log.info("merge_markers", from_entry_id=entry_2.id, to_entry_id=entry_1.id)
        await m_domain.tech_merge_markers(from_entry_id=entry_2.id, to_entry_id=entry_1.id)

    await fl_domain.tech_merge_feeds(from_feed_id=feed_2_id, to_feed_id=feed_1_id)

    await remove_feed(feed_2_id)


async def remove_feed(feed_id: uuid.UUID) -> None:
    """Remove feed and all related entries."""
    all_entries = await l_domain.get_entries_by_filter(feeds_ids=[feed_id], limit=postgresql.MAX_INTEGER)

    entries_ids = [entry.id for entry in all_entries]

    await remove_entries(entries_ids)

    await f_domain.tech_remove_feed(feed_id)


async def remove_entries(entries_ids: Iterable[uuid.UUID]) -> int:
    """Remove entries and all related markers and relations."""
    entries_to_remove = list(entries_ids)

    await m_domain.remove_markers_for_entries(entries_to_remove)
    await o_domain.remove_relations_for_entries(entries_to_remove)
    await l_domain.tech_remove_entries_by_ids(entries_to_remove)

    return len(entries_to_remove)


async def limit_entries_for_feed(feed_id: uuid.UUID, limit: int | None = None) -> None:
    """Remove oldest entries for feed to keep only `limit` entries."""
    if limit is None:
        limit = settings.max_entries_per_feed

    entries_to_remove = await l_domain.tech_get_feed_entries_tail(feed_id=feed_id, offset=limit)

    if not entries_to_remove:
        logger.info("feed_has_no_entries_tail", feed_id=feed_id, entries_limit=limit)
        return

    entries_removed = await remove_entries(entries_to_remove)

    logger.info("feed_entries_tail_removed", feed_id=feed_id, entries_limit=limit, entries_removed=entries_removed)



import uuid
from typing import Any

import pytest
import pytest_asyncio
from pytest_mock import MockerFixture

from ffun.openai.entities import KeyStatus, UserKeyInfo
from ffun.openai.keys_rotator import _get_user_key_infos
from ffun.openai.keys_statuses import statuses
from ffun.resources import domain as r_domain
from ffun.user_settings import domain as us_domain


class MockedOpenAIClient:
    def __init__(self, check_api_key: Any, request: Any) -> None:
        self.check_api_key = check_api_key
        self.request = request


async def _fake_check_api_key(api_key: str) -> KeyStatus:
    statuses.set(api_key, KeyStatus.works)
    return KeyStatus.works


# TODO: change to session scope
@pytest.fixture(autouse=True)
def mocked_openai_client(mocker: MockerFixture) -> MockedOpenAIClient:
    """Protection from calling OpenAI API during tests."""
    check_api_key = mocker.patch("ffun.openai.client.check_api_key", _fake_check_api_key)

    request = mocker.patch("ffun.openai.client.request")

    return MockedOpenAIClient(check_api_key=check_api_key, request=request)


@pytest.fixture
def openai_key() -> str:
    return uuid.uuid4().hex


@pytest_asyncio.fixture
async def five_user_key_infos(five_internal_user_ids: list[uuid.UUID]) -> list[UserKeyInfo]:
    from ffun.application.resources import Resource as AppResource
    from ffun.application.user_settings import UserSetting

    max_tokens_in_month = 1000
    used_tokens = 345

    interval_started_at = r_domain.month_interval_start()

    for user_id in five_internal_user_ids:
        await us_domain.save_setting(user_id=user_id, kind=UserSetting.openai_api_key, value=uuid.uuid4().hex)

        await us_domain.save_setting(
            user_id=user_id, kind=UserSetting.openai_max_tokens_in_month, value=max_tokens_in_month
        )

        await us_domain.save_setting(user_id=user_id, kind=UserSetting.openai_process_entries_not_older_than, value=3)

        await r_domain.try_to_reserve(
            user_id=user_id,
            kind=AppResource.openai_tokens,
            interval_started_at=interval_started_at,
            amount=used_tokens,
            limit=max_tokens_in_month,
        )

        await r_domain.convert_reserved_to_used(
            user_id=user_id,
            kind=AppResource.openai_tokens,
            interval_started_at=interval_started_at,
            used=used_tokens,
            reserved=used_tokens,
        )

    return await _get_user_key_infos(five_internal_user_ids, interval_started_at)



import datetime
import uuid

import pydantic


class BaseRule(pydantic.BaseModel):
    tags: set[int]
    score: int


class Rule(BaseRule):
    id: uuid.UUID
    user_id: uuid.UUID
    created_at: datetime.datetime
    updated_at: datetime.datetime



import datetime
import uuid
from itertools import chain

import pytest
import pytest_asyncio

from ffun.core import utils
from ffun.core.postgresql import execute
from ffun.core.tests.helpers import TableSizeDelta, TableSizeNotChanged, assert_times_is_near
from ffun.feeds import domain as f_domain
from ffun.feeds.tests import make as f_make
from ffun.library import errors
from ffun.library.domain import get_entry
from ffun.library.entities import Entry
from ffun.library.operations import (
    all_entries_iterator,
    catalog_entries,
    check_stored_entries_by_external_ids,
    get_entries_after_pointer,
    get_entries_by_filter,
    get_entries_by_ids,
    tech_get_feed_entries_tail,
    tech_move_entry,
    tech_remove_entries_by_ids,
    update_external_url,
)
from ffun.library.tests import make


class TestCatalogEntries:
    @pytest.mark.asyncio
    async def test_no_entries(self) -> None:
        async with TableSizeNotChanged("l_entries"):
            await catalog_entries([])

    @pytest.mark.asyncio
    async def test_success(self, new_entry: Entry, another_new_entry: Entry) -> None:
        entries_data = [new_entry, another_new_entry]

        async with TableSizeDelta("l_entries", delta=2):
            await catalog_entries(entries_data)

        loaded_entries = await get_entries_by_ids(ids=[new_entry.id, another_new_entry.id])

        loaded_new_entry = loaded_entries[new_entry.id]
        loaded_another_new_entry = loaded_entries[another_new_entry.id]

        assert len(loaded_entries) == 2

        assert loaded_new_entry is not None
        assert loaded_another_new_entry is not None

        assert_times_is_near(loaded_new_entry.cataloged_at, utils.now())
        assert_times_is_near(loaded_another_new_entry.cataloged_at, utils.now())

        assert loaded_new_entry == new_entry.replace(cataloged_at=loaded_new_entry.cataloged_at)
        assert loaded_another_new_entry == another_new_entry.replace(
            cataloged_at=loaded_another_new_entry.cataloged_at
        )


class TestCheckStoredEntriesByExternalIds:
    @pytest.mark.asyncio
    async def test_no_entries_stored(self, loaded_feed_id: uuid.UUID) -> None:
        entries = [make.fake_entry(loaded_feed_id) for _ in range(3)]
        external_ids = [entry.external_id for entry in entries]

        stored_entries = await check_stored_entries_by_external_ids(loaded_feed_id, external_ids)

        assert stored_entries == set()

    @pytest.mark.asyncio
    async def test_all_entries_stored(self, loaded_feed_id: uuid.UUID) -> None:
        entries = await make.n_entries(loaded_feed_id, n=3)
        external_ids = {entry.external_id for entry in entries.values()}

        stored_entries = await check_stored_entries_by_external_ids(loaded_feed_id, list(external_ids))

        assert stored_entries == external_ids

    @pytest.mark.asyncio
    async def test_some_entries_stored(self, loaded_feed_id: uuid.UUID) -> None:
        new_entries = [make.fake_entry(loaded_feed_id) for _ in range(3)]
        saved_entries = await make.n_entries(loaded_feed_id, n=2)
        external_ids = [entry.external_id for entry in new_entries] + [
            entry.external_id for entry in saved_entries.values()
        ]

        stored_entries = await check_stored_entries_by_external_ids(loaded_feed_id, external_ids)

        assert stored_entries == set(entry.external_id for entry in saved_entries.values())


class TestGetEntriesByIds:
    @pytest.mark.asyncio
    async def test_no_entries(self) -> None:
        entries = await get_entries_by_ids(ids=[])
        assert entries == {}

    @pytest.mark.asyncio
    async def test_success(self, loaded_feed_id: uuid.UUID, another_loaded_feed_id: uuid.UUID) -> None:
        entries = await make.n_entries(loaded_feed_id, n=3)
        another_entries = await make.n_entries(another_loaded_feed_id, n=3)

        entries_list = list(entries.values())
        another_entries_list = list(another_entries.values())

        entries_to_load = [*entries_list[:2], *another_entries_list[:2]]

        loaded_entries = await get_entries_by_ids(ids=[entry.id for entry in entries_to_load])

        assert len(loaded_entries) == 4
        assert entries_to_load[0] == loaded_entries[entries_to_load[0].id]
        assert entries_to_load[1] == loaded_entries[entries_to_load[1].id]
        assert another_entries_list[0] == loaded_entries[another_entries_list[0].id]
        assert another_entries_list[1] == loaded_entries[another_entries_list[1].id]


class TestGetEntriesByFilter:
    @pytest.fixture
    def time_border(self) -> datetime.datetime:
        return utils.now() - datetime.timedelta(days=1)

    @pytest_asyncio.fixture
    async def prepared_entries(
        self, loaded_feed_id: uuid.UUID, another_loaded_feed_id: uuid.UUID, time_border: datetime.datetime
    ) -> list[Entry]:
        entries = await make.n_entries(loaded_feed_id, n=3)
        another_entries = await make.n_entries(another_loaded_feed_id, n=3)

        entries_list = list(entries.values())
        another_entries_list = list(another_entries.values())

        await execute(
            "UPDATE l_entries SET created_at = %(time_border)s WHERE id = ANY(%(ids)s)",
            {
                "time_border": time_border - datetime.timedelta(seconds=10),
                "ids": [entries_list[0].id, another_entries_list[0].id],
            },
        )

        all_entries = await get_entries_by_ids(ids=[entry.id for entry in entries_list + another_entries_list])

        all_entries_list = list(all_entries.values())
        all_entries_list.sort(key=lambda entry: entry.cataloged_at)  # type: ignore

        return all_entries_list  # type: ignore

    @pytest.mark.asyncio
    async def test_all(
        self, loaded_feed_id: uuid.UUID, another_loaded_feed_id: uuid.UUID, prepared_entries: list[Entry]
    ) -> None:
        loaded_entries = await get_entries_by_filter(feeds_ids=[loaded_feed_id, another_loaded_feed_id], limit=100)

        loaded_entries_ids = {entry.id for entry in loaded_entries}
        assert loaded_entries_ids == {entry.id for entry in prepared_entries}

    @pytest.mark.asyncio
    async def test_limit(
        self, loaded_feed_id: uuid.UUID, another_loaded_feed_id: uuid.UUID, prepared_entries: list[Entry]
    ) -> None:
        loaded_entries = await get_entries_by_filter(feeds_ids=[loaded_feed_id, another_loaded_feed_id], limit=4)

        loaded_entries_ids = {entry.id for entry in loaded_entries}
        assert loaded_entries_ids == {entry.id for entry in prepared_entries[2:]}

    @pytest.mark.asyncio
    async def test_feeds_filter(self, loaded_feed_id: uuid.UUID, prepared_entries: list[Entry]) -> None:
        loaded_entries = await get_entries_by_filter(feeds_ids=[loaded_feed_id], limit=100)

        loaded_entries_ids = {entry.id for entry in loaded_entries}
        assert loaded_entries_ids == {entry.id for entry in prepared_entries if entry.feed_id == loaded_feed_id}

    @pytest.mark.asyncio
    async def test_time_period(
        self, loaded_feed_id: uuid.UUID, another_loaded_feed_id: uuid.UUID, prepared_entries: list[Entry]
    ) -> None:
        loaded_entries = await get_entries_by_filter(
            feeds_ids=[loaded_feed_id, another_loaded_feed_id], limit=100, period=datetime.timedelta(days=1)
        )

        loaded_entries_ids = {entry.id for entry in loaded_entries}
        assert loaded_entries_ids == {entry.id for entry in prepared_entries[2:]}


class TestGetEntriesAfterPointer:
    @pytest.mark.asyncio
    async def test_no_entries(self) -> None:
        entries = await get_entries_after_pointer(created_at=utils.now(), entry_id=uuid.uuid4(), limit=100)
        assert entries == []

    @pytest.mark.asyncio
    async def test_get_some(self, loaded_feed_id: uuid.UUID) -> None:
        enries = await make.n_entries(loaded_feed_id, n=5)

        entries_list = list(enries.values())
        entries_list.sort(key=lambda entry: (entry.cataloged_at, entry.id))

        loaded_entries = await get_entries_after_pointer(
            created_at=entries_list[2].cataloged_at, entry_id=entries_list[2].id, limit=100
        )

        assert [(entry.id, entry.cataloged_at) for entry in entries_list[3:]] == loaded_entries

    @pytest.mark.asyncio
    async def test_duplicated_created_at(self, loaded_feed_id: uuid.UUID) -> None:
        entries = await make.n_entries(loaded_feed_id, n=5)

        entries_list = list(entries.values())

        base_time = entries_list[1].cataloged_at

        await execute(
            "UPDATE l_entries SET created_at = %(created_at)s WHERE id = ANY(%(ids)s)",
            {"created_at": base_time, "ids": [entry.id for entry in entries_list[1:5]]},
        )

        entries = await get_entries_by_ids(ids=[entry.id for entry in entries_list])  # type: ignore

        entries_list = list(entries.values())
        entries_list.sort(key=lambda entry: (entry.cataloged_at, entry.id))

        for i in range(1, 4):
            loaded_entries = await get_entries_after_pointer(
                created_at=base_time, entry_id=entries_list[i].id, limit=100
            )

            assert [(entry.id, entry.cataloged_at) for entry in entries_list[i + 1 :]] == loaded_entries

    @pytest.mark.asyncio
    async def test_limit(self, loaded_feed_id: uuid.UUID) -> None:
        entries = await make.n_entries(loaded_feed_id, n=5)

        entries_list = list(entries.values())

        entries_list.sort(key=lambda entry: (entry.cataloged_at, entry.id))

        loaded_entries = await get_entries_after_pointer(
            created_at=entries_list[2].cataloged_at, entry_id=entries_list[2].id, limit=2
        )

        assert [(entry.id, entry.cataloged_at) for entry in entries_list[3:5]] == loaded_entries


class TestAllEntriesIterator:
    @pytest.mark.parametrize("chunk", [1, 2, 3, 4, 5, 6, 7])
    @pytest.mark.asyncio
    async def test(self, chunk: int) -> None:
        feed_1_data = f_make.fake_feed()
        feed_1_id = await f_domain.save_feed(feed_1_data)

        feed_2_data = f_make.fake_feed()
        feed_2_id = await f_domain.save_feed(feed_2_data)

        entries_1_data = [make.fake_entry(feed_1_id) for _ in range(3)]
        entries_2_data = [make.fake_entry(feed_2_id) for _ in range(3)]

        await catalog_entries(entries_1_data)
        await catalog_entries(entries_2_data)

        ids = [(e.feed_id, e.id) for e in chain(entries_1_data, entries_2_data)]

        ids.sort()

        found_ids = [
            (entry.feed_id, entry.id)
            async for entry in all_entries_iterator(chunk=chunk)
            if (entry.feed_id, entry.id) in ids
        ]

        assert found_ids == ids


class TestUpdateExternalUrl:
    @pytest.mark.asyncio
    async def test(self, cataloged_entry: Entry, another_cataloged_entry: Entry) -> None:
        new_url = make.fake_url()

        assert cataloged_entry.external_url != new_url

        await update_external_url(cataloged_entry.id, new_url)

        loaded_entry = await get_entry(cataloged_entry.id)
        assert loaded_entry.external_url == new_url

        loaded_another_entry = await get_entry(another_cataloged_entry.id)
        assert loaded_another_entry.external_url == another_cataloged_entry.external_url


class TestTechMoveEntry:
    @pytest.mark.asyncio
    async def test_moved(
        self, loaded_feed_id: uuid.UUID, another_loaded_feed_id: uuid.UUID, cataloged_entry: Entry
    ) -> None:
        async with TableSizeNotChanged("l_entries"):
            await tech_move_entry(cataloged_entry.id, another_loaded_feed_id)

        loaded_entry = await get_entry(cataloged_entry.id)

        assert loaded_entry.feed_id == another_loaded_feed_id

        assert cataloged_entry.replace(feed_id=another_loaded_feed_id) == loaded_entry

    @pytest.mark.asyncio
    async def test_feed_has_the_same_entry(
        self, loaded_feed_id: uuid.UUID, another_loaded_feed_id: uuid.UUID, new_entry: Entry
    ) -> None:
        duplicated_entry = new_entry.replace(feed_id=another_loaded_feed_id, id=uuid.uuid4())

        await catalog_entries([new_entry, duplicated_entry])

        async with TableSizeNotChanged("l_entries"):
            with pytest.raises(errors.CanNotMoveEntryAlreadyInFeed):
                await tech_move_entry(duplicated_entry.id, loaded_feed_id)

        loaded_entry = await get_entry(new_entry.id)

        assert loaded_entry == new_entry.replace(cataloged_at=loaded_entry.cataloged_at)

        loaded_duplicated_entries = await get_entry(duplicated_entry.id)

        assert loaded_duplicated_entries == duplicated_entry.replace(
            cataloged_at=loaded_duplicated_entries.cataloged_at
        )


class TestTechRemoveEntriesByIds:
    @pytest.mark.asyncio
    async def test_removed(self, loaded_feed_id: uuid.UUID, another_loaded_feed_id: uuid.UUID) -> None:
        entries = await make.n_entries(loaded_feed_id, n=3)
        another_entries = await make.n_entries(another_loaded_feed_id, n=3)

        entries_list = list(entries.values())
        another_entries_list = list(another_entries.values())

        async with TableSizeDelta("l_entries", delta=-2):
            await tech_remove_entries_by_ids([entries_list[0].id, another_entries_list[0].id])

        loaded_entries = await get_entries_by_ids([entry.id for entry in entries_list + another_entries_list])

        assert loaded_entries[entries_list[0].id] is None
        assert loaded_entries[another_entries_list[0].id] is None

    @pytest.mark.asyncio
    async def test_no_entries(self) -> None:
        async with TableSizeNotChanged("l_entries"):
            await tech_remove_entries_by_ids([])
            await tech_remove_entries_by_ids([uuid.uuid4()])

    @pytest.mark.asyncio
    async def test_already_removed(self, loaded_feed_id: uuid.UUID) -> None:
        entries = await make.n_entries(loaded_feed_id, n=3)

        entries_list = list(entries.values())

        await tech_remove_entries_by_ids([entry.id for entry in entries_list])

        async with TableSizeNotChanged("l_entries"):
            await tech_remove_entries_by_ids([entry.id for entry in entries_list])


class TestTechGetFeedEntriesTail:
    @pytest.mark.asyncio
    async def test_nothing_to_return(self, loaded_feed_id: uuid.UUID) -> None:
        await make.n_entries(loaded_feed_id, n=5)

        ids = await tech_get_feed_entries_tail(loaded_feed_id, offset=10)

        assert ids == set()

    @pytest.mark.asyncio
    async def test_zero_head(self, loaded_feed_id: uuid.UUID) -> None:
        entries = await make.n_entries(loaded_feed_id, n=5)

        ids = await tech_get_feed_entries_tail(loaded_feed_id, offset=0)

        assert ids == set(entry.id for entry in entries.values())

    @pytest.mark.asyncio
    async def test_not_excceed_limit(self, loaded_feed_id: uuid.UUID) -> None:
        await make.n_entries(loaded_feed_id, n=5)

        ids = await tech_get_feed_entries_tail(loaded_feed_id, offset=10)

        assert ids == set()

    @pytest.mark.asyncio
    async def test_offset(self, loaded_feed_id: uuid.UUID) -> None:
        entries = await make.n_entries_list(loaded_feed_id, n=15)

        ids = await tech_get_feed_entries_tail(loaded_feed_id, offset=10)

        assert ids == set(entry.id for entry in entries[10:])



import uuid
from importlib import metadata
from typing import Any, Iterable

import fastapi
from fastapi.openapi.docs import get_swagger_ui_html
from fastapi.openapi.utils import get_openapi
from fastapi.responses import HTMLResponse, JSONResponse

from ffun.api import entities
from ffun.api.settings import settings
from ffun.auth.dependencies import User
from ffun.core import logging
from ffun.feeds import domain as f_domain
from ffun.feeds import entities as f_entities
from ffun.feeds_collections import domain as fc_domain
from ffun.feeds_discoverer import domain as fd_domain
from ffun.feeds_links import domain as fl_domain
from ffun.library import domain as l_domain
from ffun.library import entities as l_entities
from ffun.markers import domain as m_domain
from ffun.ontology import domain as o_domain
from ffun.parsers import domain as p_domain
from ffun.parsers import entities as p_entities
from ffun.resources import domain as r_domain
from ffun.scores import domain as s_domain
from ffun.scores import entities as s_entities
from ffun.user_settings import domain as us_domain

router = fastapi.APIRouter()

logger = logging.get_module_logger()


@router.post("/api/error")
async def api_error() -> None:
    raise Exception("test_error")


@router.post("/api/get-feeds")
async def api_get_feeds(request: entities.GetFeedsRequest, user: User) -> entities.GetFeedsResponse:
    linked_feeds = await fl_domain.get_linked_feeds(user.id)

    feeds_to_links = {link.feed_id: link for link in linked_feeds}

    feeds = await f_domain.get_feeds(ids=list(feeds_to_links.keys()))

    return entities.GetFeedsResponse(
        feeds=[entities.Feed.from_internal(feed, link=feeds_to_links[feed.id]) for feed in feeds]
    )


async def _external_entries(  # pylint: disable=R0914
    entries: Iterable[l_entities.Entry], with_body: bool, user_id: uuid.UUID
) -> list[entities.Entry]:
    entries_ids = [entry.id for entry in entries]

    tags = await o_domain.get_tags_for_entries(entries_ids)

    tags_ids = await o_domain.get_tags_ids_for_entries(entries_ids)

    markers = await m_domain.get_markers(user_id=user_id, entries_ids=entries_ids)

    rules = await s_domain.get_rules(user_id)

    external_entries = []

    for entry in entries:
        score, contributions_by_ids = s_domain.get_score_contributions(rules, tags_ids.get(entry.id, set()))

        tags_mapping = await o_domain.get_tags_by_ids(contributions_by_ids.keys())

        contributions_by_str = {
            tags_mapping[tag_id]: contribution for tag_id, contribution in contributions_by_ids.items()
        }

        external_markers = [entities.Marker.from_internal(marker) for marker in markers.get(entry.id, ())]

        external_entry = entities.Entry.from_internal(
            entry=entry,
            tags=tags.get(entry.id, ()),
            markers=external_markers,
            score=score,
            score_contributions=contributions_by_str,
            with_body=with_body,
        )

        external_entries.append(external_entry)

    external_entries.sort(key=lambda entry: entry.score, reverse=True)

    return external_entries


@router.post("/api/get-last-entries")
async def api_get_last_entries(request: entities.GetLastEntriesRequest, user: User) -> entities.GetLastEntriesResponse:
    linked_feeds = await fl_domain.get_linked_feeds(user.id)

    linked_feeds_ids = [link.feed_id for link in linked_feeds]

    entries = await l_domain.get_entries_by_filter(
        feeds_ids=linked_feeds_ids, period=request.period, limit=settings.max_returned_entries
    )

    external_entries = await _external_entries(entries, with_body=False, user_id=user.id)

    return entities.GetLastEntriesResponse(entries=external_entries)


@router.post("/api/get-entries-by-ids")
async def api_get_entries_by_ids(
    request: entities.GetEntriesByIdsRequest, user: User
) -> entities.GetEntriesByIdsResponse:
    # TODO: check if belongs to user

    if len(request.ids) > 10:
        # TODO: better error processing
        raise fastapi.HTTPException(status_code=400, detail="Too many ids")

    entries = await l_domain.get_entries_by_ids(request.ids)

    found_entries = [entry for entry in entries.values() if entry is not None]

    external_entries = await _external_entries(found_entries, with_body=True, user_id=user.id)

    return entities.GetEntriesByIdsResponse(entries=external_entries)


@router.post("/api/create-or-update-rule")
async def api_create_or_update_rule(
    request: entities.CreateOrUpdateRuleRequest, user: User
) -> entities.CreateOrUpdateRuleResponse:
    tags_ids = await o_domain.get_ids_by_uids(request.tags)

    await s_domain.create_or_update_rule(user_id=user.id, tags=set(tags_ids.values()), score=request.score)

    return entities.CreateOrUpdateRuleResponse()


@router.post("/api/delete-rule")
async def api_delete_rule(request: entities.DeleteRuleRequest, user: User) -> entities.DeleteRuleResponse:
    await s_domain.delete_rule(user_id=user.id, rule_id=request.id)

    return entities.DeleteRuleResponse()


@router.post("/api/update-rule")
async def api_update_rule(request: entities.UpdateRuleRequest, user: User) -> entities.UpdateRuleResponse:
    tags_ids = await o_domain.get_ids_by_uids(request.tags)

    await s_domain.update_rule(user_id=user.id, rule_id=request.id, score=request.score, tags=tags_ids.values())

    return entities.UpdateRuleResponse()


async def _prepare_rules(rules: Iterable[s_entities.Rule]) -> list[entities.Rule]:
    all_tags = set()

    for rule in rules:
        all_tags.update(rule.tags)

    tags_mapping = await o_domain.get_tags_by_ids(all_tags)

    external_rules = [entities.Rule.from_internal(rule=rule, tags_mapping=tags_mapping) for rule in rules]

    return external_rules


@router.post("/api/get-rules")
async def api_get_rules(request: entities.GetRulesRequest, user: User) -> entities.GetRulesResponse:
    rules = await s_domain.get_rules(user_id=user.id)

    external_rules = await _prepare_rules(rules)

    return entities.GetRulesResponse(rules=external_rules)


@router.post("/api/get-score-details")
async def api_get_score_details(
    request: entities.GetScoreDetailsRequest, user: User
) -> entities.GetScoreDetailsResponse:
    entry_id = request.entryId

    rules = await s_domain.get_rules(user.id)

    tags_ids = await o_domain.get_tags_ids_for_entries([entry_id])

    rules = s_domain.get_score_rules(rules, tags_ids.get(entry_id, set()))

    external_rules = await _prepare_rules(rules)

    return entities.GetScoreDetailsResponse(rules=external_rules)


@router.post("/api/set-marker")
async def api_set_marker(request: entities.SetMarkerRequest, user: User) -> entities.SetMarkerResponse:
    await m_domain.set_marker(user_id=user.id, entry_id=request.entryId, marker=request.marker.to_internal())

    return entities.SetMarkerResponse()


@router.post("/api/remove-marker")
async def api_remove_marker(request: entities.RemoveMarkerRequest, user: User) -> entities.RemoveMarkerResponse:
    await m_domain.remove_marker(user_id=user.id, entry_id=request.entryId, marker=request.marker.to_internal())

    return entities.RemoveMarkerResponse()


@router.post("/api/discover-feeds")
async def api_discover_feeds(request: entities.DiscoverFeedsRequest, user: User) -> entities.DiscoverFeedsResponse:
    feeds = await fd_domain.discover(url=request.url)

    for feed in feeds:
        # TODO: should we limit entities number there?
        feed.entries = feed.entries[:3]

    external_feeds = [entities.FeedInfo.from_internal(feed) for feed in feeds]

    return entities.DiscoverFeedsResponse(feeds=external_feeds)


async def _add_feeds(feed_infos: list[p_entities.FeedInfo], user: User) -> None:
    feeds = [
        f_entities.Feed(id=uuid.uuid4(), url=feed_info.url, title=feed_info.title, description=feed_info.description)
        for feed_info in feed_infos
    ]

    real_feeds_ids = await f_domain.save_feeds(feeds)

    for feed_id in real_feeds_ids:
        await fl_domain.add_link(user_id=user.id, feed_id=feed_id)


@router.post("/api/add-feed")
async def api_add_feed(request: entities.AddFeedRequest, user: User) -> entities.AddFeedResponse:
    feed_info = await fd_domain.check_if_feed(url=request.url)

    if feed_info is None:
        raise fastapi.HTTPException(status_code=400, detail="Not a feed")

    await _add_feeds([feed_info], user)

    return entities.AddFeedResponse()


@router.post("/api/add-opml")
async def api_add_opml(request: entities.AddOpmlRequest, user: User) -> entities.AddOpmlResponse:
    feed_infos = p_domain.parse_opml(request.content)

    await _add_feeds(feed_infos, user)

    return entities.AddOpmlResponse()


@router.post("/api/unsubscribe")
async def api_unsubscribe(request: entities.UnsubscribeRequest, user: User) -> entities.UnsubscribeResponse:
    await fl_domain.remove_link(user_id=user.id, feed_id=request.feedId)

    return entities.UnsubscribeResponse()


@router.post("/api/get-feeds-collections")
async def api_get_feeds_collections(
    request: entities.GetFeedsCollectionsRequest, user: User
) -> entities.GetFeedsCollectionsResponse:
    collections = list(fc_domain.get_collections())

    return entities.GetFeedsCollectionsResponse(collections=collections)


@router.post("/api/subscribe-to-feeds-collections")
async def api_subscribe_to_feeds_collections(
    request: entities.SubscribeToFeedsCollectionsRequest, user: User
) -> entities.SubscribeToFeedsCollectionsResponse:
    feeds = []

    for collection in request.collections:
        feed_urls = fc_domain.get_feeds_for_collecton(collection)

        for feed_url in feed_urls:
            feeds.append(p_entities.FeedInfo(url=feed_url, title="unknown", description="unknown", entries=[]))

    await _add_feeds(feeds, user)

    return entities.SubscribeToFeedsCollectionsResponse()


@router.post("/api/get-tags-info")
async def api_get_tags_info(request: entities.GetTagsInfoRequest, user: User) -> entities.GetTagsInfoResponse:
    tags_ids = await o_domain.get_ids_by_uids(request.uids)

    info = await o_domain.get_tags_info(tags_ids.values())

    tags_info = {}

    for uid in request.uids:
        tags_info[uid] = entities.TagInfo.from_internal(info[tags_ids[uid]], uid)

    return entities.GetTagsInfoResponse(tags=tags_info)


@router.post("/api/get-resource-history")
async def api_get_resource_history(
    request: entities.GetResourceHistoryRequest, user: User
) -> entities.GetResourceHistoryResponse:
    history = await r_domain.load_resource_history(user_id=user.id, kind=request.kind.to_internal())

    return entities.GetResourceHistoryResponse(
        history=[entities.ResourceHistoryRecord.from_internal(resource) for resource in history]
    )


@router.post("/api/get-info")
async def api_get_info(request: entities.GetInfoRequest, user: User) -> entities.GetInfoResponse:
    return entities.GetInfoResponse(userId=user.id)


###############
# user settings
###############


@router.post("/api/get-user-settings")
async def api_get_user_settings(
    request: entities.GetUserSettingsRequest, user: User
) -> entities.GetUserSettingsResponse:
    from ffun.application.user_settings import UserSetting

    values = await us_domain.load_settings(user_id=user.id, kinds=[int(kind) for kind in UserSetting])

    result_values = []

    for kind, value in values.items():
        result_values.append(entities.UserSetting.from_internal(kind, value))

    return entities.GetUserSettingsResponse(settings=result_values)


@router.post("/api/set-user-setting")
async def api_set_user_setting(request: entities.SetUserSettingRequest, user: User) -> entities.SetUserSettingResponse:
    await us_domain.save_setting(user_id=user.id, kind=request.kind.to_internal(), value=request.value)

    return entities.SetUserSettingResponse()


#######################
# Swagger documentation
#######################

swagger_ui_api_parameters: dict[str, Any] = {
    "defaultModelsExpandDepth": -1,
    "defaultModelExpandDepth": 10,
}


swagger_title = "Feeds Fun API"


swagger_description = """
Greetings!

Welcome to the documentation for the Feeds Fun API.

Please note that the project is currently in its early stages of development, and as such, the API may undergo \
significant changes. We appreciate your understanding and patience during this phase.

At present, our **documentation does not include information regarding authentication**. If you wish to utilize API \
from [feeds.fun](https://feeds.fun), we recommend referring to the [supertokens](https://supertokens.com/) \
documentation for guidance on this matter. We'll improve this aspect of the documentation in the future.

For additional resources, please visit the following links:

- [Feeds Fun Website](https://feeds.fun)
- [GitHub Repository](https://github.com/Tiendil/feeds.fun)

Thank you for your interest in the Feeds Fun API. We look forward to your contributions and feedback.

"""


@router.get("/api/openapi.json", include_in_schema=False)
async def openapi(request: fastapi.Request) -> JSONResponse:
    content = get_openapi(
        title="Feeds Fun API",
        version=metadata.version("ffun"),
        description=swagger_description,
        routes=request.app.routes,
    )

    return JSONResponse(content=content)


@router.get("/api/docs", include_in_schema=False)
async def docs(request: fastapi.Request) -> HTMLResponse:
    openapi_url = request.scope.get("root_path", "") + "/api/openapi.json"

    return get_swagger_ui_html(
        openapi_url=openapi_url, title=swagger_title, swagger_ui_parameters=swagger_ui_api_parameters
    )



import datetime
from typing import Any, Iterable

import feedparser

from ffun.core import logging
from ffun.domain import urls
from ffun.parsers.entities import EntryInfo, FeedInfo

logger = logging.get_module_logger()


def _parse_tags(tags: Iterable[dict[str, Any]]) -> set[str]:
    result = set()

    for tag in tags:
        if tag.get("label") is not None:
            result.add(tag["label"])

        elif tag.get("term") is not None:
            result.add(tag["term"])

    return result


def _should_skip(entry: Any) -> bool:
    if entry.get("link") is None:
        logger.warning("feed_does_not_has_link_field")
        return True

    return False


def _extract_published_at(entry: Any) -> datetime.datetime:
    published_at = entry.get("published_parsed")

    if published_at is not None:
        return datetime.datetime(*published_at[:6])

    return datetime.datetime.now()


# not the best solution, but it works for now
# TODO: we should use more formal way to detect uniqueness of entry
# TODO: external id must be normalized
def _extract_external_id(entry: Any) -> str:
    return entry.get("link")  # type: ignore


def _extract_external_url(entry: Any, original_url: str) -> str | None:
    url = entry.get("link")

    return urls.normalize_external_url(url, original_url)


def parse_feed(content: str, original_url: str) -> FeedInfo | None:
    channel = feedparser.parse(content)

    if getattr(channel, "version", "") == "" and not channel.entries:
        return None

    feed_info = FeedInfo(
        url=original_url,
        title=channel.feed.get("title", ""),
        description=channel.feed.get("description", ""),
        entries=[],
    )

    for entry in channel.entries:
        # TODO: remove all tags from title
        # TODO: extract tags from <category> tag

        if _should_skip(entry):
            continue

        published_at = _extract_published_at(entry)

        feed_info.entries.append(
            EntryInfo(
                title=entry.get("title", ""),
                body=entry.get("description", ""),
                external_id=_extract_external_id(entry),
                external_url=_extract_external_url(entry, original_url),
                external_tags=_parse_tags(entry.get("tags", ())),
                published_at=published_at,
            )
        )

    return feed_info



import pydantic
import pydantic_settings

from ffun.core.settings import BaseSettings


class BaseProcessor(pydantic.BaseModel):
    enabled: bool = False
    workers: int = 1


class DomainProcessor(BaseProcessor):
    pass


class NativeTagsProcessor(BaseProcessor):
    pass


class UpperCaseTitleProcessor(BaseProcessor):
    pass


class OpenAIChat35Processor(BaseProcessor):
    model: str = "gpt-3.5-turbo-0125"


class OpenAIChat35FunctionsProcessor(BaseProcessor):
    model: str = "gpt-3.5-turbo-0125"


class Settings(BaseSettings):
    domain_processor: DomainProcessor = DomainProcessor(enabled=True)
    native_tags_processor: NativeTagsProcessor = NativeTagsProcessor(enabled=True)
    upper_case_title_processor: UpperCaseTitleProcessor = UpperCaseTitleProcessor(enabled=True)

    openai_chat_35_processor: OpenAIChat35Processor = OpenAIChat35Processor()
    openai_chat_35_functions_processor: OpenAIChat35FunctionsProcessor = OpenAIChat35FunctionsProcessor()

    model_config = pydantic_settings.SettingsConfigDict(env_prefix="FFUN_LIBRARIAN_")


settings = Settings()



import uuid

from ffun.core.postgresql import ExecuteType, run_in_transaction
from ffun.feeds_links import operations

add_link = operations.add_link
remove_link = operations.remove_link
get_linked_feeds = operations.get_linked_feeds
get_linked_users = operations.get_linked_users
has_linked_users = operations.has_linked_users


@run_in_transaction
async def tech_merge_feeds(execute: ExecuteType, from_feed_id: uuid.UUID, to_feed_id: uuid.UUID) -> None:
    await operations.tech_merge_feeds(execute, from_feed_id=from_feed_id, to_feed_id=to_feed_id)



import xml.etree.ElementTree as ET  # noqa
from typing import Generator

from ffun.parsers.entities import FeedInfo


def _extract_body(data: str) -> ET.Element:
    root = ET.fromstring(data)  # noqa

    if root.tag != "opml":
        raise NotImplementedError(f"Unknown root tag: {root.tag}")

    head, body = root

    if head.tag != "head":
        raise NotImplementedError(f"Unknown root tag: {head.tag}")

    if body.tag != "body":
        raise NotImplementedError(f"Unknown root tag: {body.tag}")

    return body


def extract_feeds(data: str) -> list[FeedInfo]:
    feeds: list[FeedInfo] = []

    body = _extract_body(data)

    repeated_feeds = extract_feeds_records(body)

    for new_feed in repeated_feeds:
        for saved_feed in feeds:
            if new_feed.url == saved_feed.url:
                break
        else:
            feeds.append(new_feed)

    return feeds


def extract_feeds_records(body: ET.Element) -> Generator[FeedInfo, None, None]:
    for outline in body:
        if outline.attrib.get("type") == "rss":
            yield FeedInfo(
                url=outline.attrib["xmlUrl"],
                title=outline.attrib.get("title", ""),
                description="",
                entries=[],
            )
            continue

        yield from extract_feeds_records(outline)



collection = {
    "http://roguelikedeveloper.blogspot.com/feeds/posts/default",
    "http://engineering.riotgames.com/rss.xml",
    "http://blog.ihobo.com/atom.xml",
    "http://www.raphkoster.com/feed/",
    "https://www.deconstructoroffun.com/blog?format=rss",
    "https://heredragonsabound.blogspot.com/feeds/posts/default",
    "https://godotengine.org/rss.xml",
    "http://www.filfre.net/feed/",
    "https://pvigier.github.io/rss.xml",
    "http://www.reddit.com/r/proceduralgeneration/.rss",
    "https://www.gridsagegames.com/blog",
    "http://www.gamedevblog.com/index.rdf",
    "http://fabiensanglard.net/rss.xml",
    "http://www.indiegames.com/blog/atom.xml",
    "http://www.zenofdesign.com/feed/",
    "https://lostgarden.home.blog/feed/",
    "http://www.reddit.com/r/gameai/.rss",
    "http://greyaliengames.com/blog/feed/",
    "http://www.reddit.com/r/roguelikedev/.rss",
    "https://heavens-vault-game.tumblr.com/rss",
    "https://tech.spaceapegames.com/feed/",
    "https://www.boristhebrave.com/feed/",
    "http://0fps.wordpress.com/feed/",
    "http://www.kloonigames.com/blog/feed/",
    "http://www.the-working-man.org/feeds/posts/default",
    "http://andrewfray.wordpress.com/feed/",
    "http://simblob.blogspot.com/feeds/posts/default",
    "https://www.reddit.com/r/alife/.rss",
    "http://www.alanzucconi.com/feed/",
}



import re
import unicodedata
from urllib.parse import quote_plus, unquote

from furl import furl
from orderedmultidict import omdict

from ffun.core import logging

logger = logging.get_module_logger()


RE_SCHEMA = re.compile(r"^(\w+):")


# TODO: maybe switch to https://github.com/tktech/can_ada


# TODO: add tests
# is required for correct parsing by furl
# will be removed before returning result
def _fake_schema_for_url(url: str) -> str:
    if url.startswith("//"):
        return url

    if RE_SCHEMA.match(url):
        return url

    # TODO: this logic is required only for normalize_classic_url and has wrong behavior for top-level domains
    #       like localhost or any other one-part domain that user will define locally
    if "." not in url.split("/")[0]:
        # if there is no domain, just return the url
        return url

    return f"//{url}"


def normalize_classic_url(url: str, original_url: str) -> str | None:
    url = _fake_schema_for_url(url)
    original_url = _fake_schema_for_url(original_url)

    try:
        external_url = furl(url)
    except ValueError as e:
        error = str(e)

        logger.warning("invalid_url", url=url, error=error)

        if "Invalid port" in error:
            return None

        raise

    f_original_url = furl(original_url)

    if not external_url.scheme:
        external_url.set(scheme=f_original_url.scheme)

    if not external_url.netloc:
        external_url.set(netloc=f_original_url.netloc)

    result_url = str(external_url)

    if result_url.startswith("//"):
        result_url = result_url[2:]

    return result_url


def is_magnetic_url(url: str) -> bool:
    return url.startswith("magnet:")


def normalize_magnetic_url(url: str) -> str:
    return url


def normalize_external_url(url: str, original_url: str) -> str | None:
    if is_magnetic_url(url):
        return normalize_magnetic_url(url)

    return normalize_classic_url(url, original_url)


def url_to_uid(url: str) -> str:
    # The goal of this function is to detect urls that most likely (99.(9)%) point to the same resource
    # It normalizes and simplifies an url according to heuristics
    # I.e. there is a small possibility that two different urls will be normalized to the same uid
    #
    # For example, http://example.com/ and http://example.com can return different data
    # In reality it is mostly imposible and is a signe of bug or hacking on the side of third-party service
    # => we could remove schema from the resulted uid
    #
    # Some normalization rules are based on personal taste,
    # for example, there are multiple ways to encode a url or to normalize unicode
    #
    # The rules are based on the next heuristics:
    #
    # - readability are better than technical representation
    # - it is ok to loss some corner urls forms, unless there will be an explicit request to support them

    normalized_url = _fake_schema_for_url(url.lower().strip())

    normalized_url = unquote(normalized_url)  # unquote all
    normalized_url = quote_plus(normalized_url)  # quoute with replacing spaces with pluses
    normalized_url = unquote(normalized_url)  # unquote all again, but keep pluses

    normalized_url = unicodedata.normalize("NFC", normalized_url)

    url_object = furl(normalized_url)

    url_object.scheme = None
    url_object.port = None
    url_object.fragment = None

    url_object.query.params = omdict(sorted(url_object.query.params.allitems()))

    # Attention: we must not remove username:password from the url
    #            because it will create a vector for hacking by accessing private news of other users

    path = str(url_object.path)

    while "//" in path:
        path = path.replace("//", "/")

    if path and path[-1] == "/":
        path = path[:-1]

    url_object.path = path

    # unquote again because furl will quote all parts of the url
    resulted_url = unquote(str(url_object))

    # furl adds // at the beginning of the url in some cases
    if resulted_url.startswith("//"):
        resulted_url = resulted_url[2:]

    return resulted_url



import uuid
from itertools import chain

import pytest

from ffun.feeds import domain as f_domain
from ffun.feeds import errors as f_errors
from ffun.feeds_links import domain as fl_domain
from ffun.library import domain as l_domain
from ffun.library.tests import make as l_make
from ffun.markers import domain as m_domains
from ffun.markers.entities import Marker
from ffun.meta.domain import limit_entries_for_feed, merge_feeds, remove_entries, remove_feed
from ffun.ontology import domain as o_domain
from ffun.ontology.entities import ProcessorTag
from ffun.users.tests import make as u_make


class TestRemoveFeed:
    @pytest.mark.asyncio
    async def test_no_feed(self) -> None:
        await remove_feed(uuid.uuid4())

    @pytest.mark.asyncio
    async def test_success(
        self,
        loaded_feed_id: uuid.UUID,
        another_loaded_feed_id: uuid.UUID,
        fake_processor_id: int,
        another_fake_processor_id: int,
        three_processor_tags: tuple[ProcessorTag, ProcessorTag, ProcessorTag],
    ) -> None:
        entries = await l_make.n_entries_list(loaded_feed_id, 3)
        another_entries = await l_make.n_entries_list(another_loaded_feed_id, 3)

        tag_a, tag_b, tag_c = three_processor_tags

        # fill feed 1
        await o_domain.apply_tags_to_entry(entry_id=entries[0].id, processor_id=fake_processor_id, tags=[tag_a])

        await o_domain.apply_tags_to_entry(entry_id=entries[1].id, processor_id=fake_processor_id, tags=[tag_a])

        await o_domain.apply_tags_to_entry(
            entry_id=entries[1].id, processor_id=another_fake_processor_id, tags=[tag_b]
        )

        # fill feed 2
        await o_domain.apply_tags_to_entry(
            entry_id=another_entries[1].id, processor_id=fake_processor_id, tags=[tag_a]
        )

        await o_domain.apply_tags_to_entry(
            entry_id=another_entries[1].id, processor_id=another_fake_processor_id, tags=[tag_b]
        )

        await o_domain.apply_tags_to_entry(
            entry_id=another_entries[2].id, processor_id=another_fake_processor_id, tags=[tag_c]
        )

        await remove_feed(loaded_feed_id)

        loaded_entries = await l_domain.get_entries_by_ids(
            [entry.id for entry in entries] + [entry.id for entry in another_entries]
        )

        assert loaded_entries == {
            entries[0].id: None,
            entries[1].id: None,
            entries[2].id: None,
            another_entries[0].id: another_entries[0],
            another_entries[1].id: another_entries[1],
            another_entries[2].id: another_entries[2],
        }

    @pytest.mark.asyncio
    async def test_remove_markers(self, loaded_feed_id: uuid.UUID, another_loaded_feed_id: uuid.UUID) -> None:
        entries = await l_make.n_entries_list(loaded_feed_id, 3)

        another_entries = await l_make.n_entries_list(another_loaded_feed_id, 3)

        user_a, user_b, user_c = await u_make.n_users(3)

        await m_domains.set_marker(user_a, Marker.read, entries[0].id)
        await m_domains.set_marker(user_b, Marker.read, entries[1].id)
        await m_domains.set_marker(user_b, Marker.read, entries[2].id)

        await m_domains.set_marker(user_b, Marker.read, another_entries[0].id)
        await m_domains.set_marker(user_c, Marker.read, another_entries[1].id)
        await m_domains.set_marker(user_c, Marker.read, another_entries[2].id)

        await remove_feed(loaded_feed_id)

        markers_a = await m_domains.get_markers(user_a, [entry.id for entry in chain(entries, another_entries)])
        markers_b = await m_domains.get_markers(user_b, [entry.id for entry in chain(entries, another_entries)])
        markers_c = await m_domains.get_markers(user_c, [entry.id for entry in chain(entries, another_entries)])

        assert markers_a == {}
        assert markers_b == {another_entries[0].id: {Marker.read}}
        assert markers_c == {another_entries[1].id: {Marker.read}, another_entries[2].id: {Marker.read}}


class TestMergeFeeds:
    @pytest.mark.asyncio
    async def test_no_entries_in_source_feed(
        self, loaded_feed_id: uuid.UUID, another_loaded_feed_id: uuid.UUID
    ) -> None:
        await merge_feeds(loaded_feed_id, another_loaded_feed_id)

    @pytest.mark.asyncio
    async def test_move_entries_to_an_empty_feed(
        self,
        loaded_feed_id: uuid.UUID,
        another_loaded_feed_id: uuid.UUID,
        fake_processor_id: int,
        another_fake_processor_id: int,
        three_processor_tags: tuple[ProcessorTag, ProcessorTag, ProcessorTag],
    ) -> None:
        entries = await l_make.n_entries_list(loaded_feed_id, 3)

        tag_a, tag_b, tag_c = three_processor_tags

        await o_domain.apply_tags_to_entry(entry_id=entries[0].id, processor_id=fake_processor_id, tags=[tag_a])

        await o_domain.apply_tags_to_entry(entry_id=entries[1].id, processor_id=fake_processor_id, tags=[tag_a])

        await o_domain.apply_tags_to_entry(
            entry_id=entries[1].id, processor_id=another_fake_processor_id, tags=[tag_b]
        )

        await merge_feeds(another_loaded_feed_id, loaded_feed_id)

        with pytest.raises(f_errors.NoFeedFound):
            await f_domain.get_feed(loaded_feed_id)

        loaded_entries = await l_domain.get_entries_by_filter(feeds_ids=[another_loaded_feed_id], limit=100)

        assert loaded_entries == [entry.replace(feed_id=another_loaded_feed_id) for entry in entries]

        tags = await o_domain.get_tags_for_entries([entry.id for entry in loaded_entries])

        assert tags == {
            entries[0].id: {tag_a.raw_uid},
            entries[1].id: {tag_a.raw_uid, tag_b.raw_uid},
            entries[2].id: set(),
        }

    @pytest.mark.asyncio
    async def test_merge_entries(
        self,
        loaded_feed_id: uuid.UUID,
        another_loaded_feed_id: uuid.UUID,
        fake_processor_id: int,
        another_fake_processor_id: int,
        three_processor_tags: tuple[ProcessorTag, ProcessorTag, ProcessorTag],
    ) -> None:
        entries = await l_make.n_entries_list(loaded_feed_id, 3)

        another_entries_to_save = [entry.replace(feed_id=another_loaded_feed_id, id=uuid.uuid4()) for entry in entries]
        await l_domain.catalog_entries(reversed(another_entries_to_save))

        another_entries = await l_domain.get_entries_by_filter(feeds_ids=[another_loaded_feed_id], limit=100)

        tag_a, tag_b, tag_c = three_processor_tags

        # fill feed 1
        await o_domain.apply_tags_to_entry(entry_id=entries[0].id, processor_id=fake_processor_id, tags=[tag_a])

        await o_domain.apply_tags_to_entry(entry_id=entries[1].id, processor_id=fake_processor_id, tags=[tag_a])

        await o_domain.apply_tags_to_entry(
            entry_id=entries[1].id, processor_id=another_fake_processor_id, tags=[tag_b]
        )

        # fill feed 2
        await o_domain.apply_tags_to_entry(
            entry_id=another_entries[1].id, processor_id=fake_processor_id, tags=[tag_c]
        )

        await o_domain.apply_tags_to_entry(
            entry_id=another_entries[1].id, processor_id=another_fake_processor_id, tags=[tag_b]
        )

        await o_domain.apply_tags_to_entry(
            entry_id=another_entries[2].id, processor_id=another_fake_processor_id, tags=[tag_c]
        )

        await merge_feeds(another_loaded_feed_id, loaded_feed_id)

        with pytest.raises(f_errors.NoFeedFound):
            await f_domain.get_feed(loaded_feed_id)

        loaded_entries = await l_domain.get_entries_by_filter(feeds_ids=[another_loaded_feed_id], limit=100)

        assert loaded_entries == another_entries
        assert loaded_entries == [
            entry.replace(
                feed_id=another_loaded_feed_id, id=another_entries[i].id, cataloged_at=another_entries[i].cataloged_at
            )
            for i, entry in enumerate(entries)
        ]

        tags = await o_domain.get_tags_for_entries([entry.id for entry in loaded_entries])

        assert tags == {
            another_entries[0].id: {tag_a.raw_uid},
            another_entries[1].id: {tag_a.raw_uid, tag_b.raw_uid, tag_c.raw_uid},
            another_entries[2].id: {tag_c.raw_uid},
        }

    @pytest.mark.asyncio
    async def test_move_and_merge(
        self,
        loaded_feed_id: uuid.UUID,
        another_loaded_feed_id: uuid.UUID,
        fake_processor_id: int,
        another_fake_processor_id: int,
        three_processor_tags: tuple[ProcessorTag, ProcessorTag, ProcessorTag],
    ) -> None:
        entries = await l_make.n_entries_list(loaded_feed_id, 3)

        another_entries_to_save = [
            entry.replace(feed_id=another_loaded_feed_id, id=uuid.uuid4()) for entry in entries[1:]
        ]
        await l_domain.catalog_entries(reversed(another_entries_to_save))

        another_entries = await l_domain.get_entries_by_filter(feeds_ids=[another_loaded_feed_id], limit=100)

        tag_a, tag_b, tag_c = three_processor_tags

        # fill feed 1
        await o_domain.apply_tags_to_entry(entry_id=entries[0].id, processor_id=fake_processor_id, tags=[tag_a])

        await o_domain.apply_tags_to_entry(entry_id=entries[1].id, processor_id=fake_processor_id, tags=[tag_a])

        await o_domain.apply_tags_to_entry(
            entry_id=entries[1].id, processor_id=another_fake_processor_id, tags=[tag_b]
        )

        # fill feed 2
        await o_domain.apply_tags_to_entry(
            entry_id=another_entries[0].id, processor_id=fake_processor_id, tags=[tag_c]
        )

        await o_domain.apply_tags_to_entry(
            entry_id=another_entries[0].id, processor_id=another_fake_processor_id, tags=[tag_b]
        )

        await o_domain.apply_tags_to_entry(
            entry_id=another_entries[1].id, processor_id=another_fake_processor_id, tags=[tag_c]
        )

        await merge_feeds(another_loaded_feed_id, loaded_feed_id)

        with pytest.raises(f_errors.NoFeedFound):
            await f_domain.get_feed(loaded_feed_id)

        loaded_entries = await l_domain.get_entries_by_filter(feeds_ids=[another_loaded_feed_id], limit=100)

        assert loaded_entries[-1] == entries[0].replace(feed_id=another_loaded_feed_id)
        assert loaded_entries[:-1] == another_entries

        tags = await o_domain.get_tags_for_entries([entry.id for entry in loaded_entries])

        assert tags == {
            entries[0].id: {tag_a.raw_uid},
            another_entries[0].id: {tag_a.raw_uid, tag_b.raw_uid, tag_c.raw_uid},
            another_entries[1].id: {tag_c.raw_uid},
        }

    @pytest.mark.asyncio
    async def test_merge_feed_links(self, loaded_feed_id: uuid.UUID, another_loaded_feed_id: uuid.UUID) -> None:
        user_a, user_b = await u_make.n_users(2)

        await fl_domain.add_link(user_a, another_loaded_feed_id)
        await fl_domain.add_link(user_b, loaded_feed_id)
        await fl_domain.add_link(user_b, another_loaded_feed_id)

        await merge_feeds(loaded_feed_id, another_loaded_feed_id)

        links_a = await fl_domain.get_linked_feeds(user_a)
        links_b = await fl_domain.get_linked_feeds(user_b)

        assert {link.feed_id for link in links_a} == {loaded_feed_id}
        assert {link.feed_id for link in links_b} == {loaded_feed_id}

    @pytest.mark.asyncio
    async def test_merge_markers(self, loaded_feed_id: uuid.UUID, another_loaded_feed_id: uuid.UUID) -> None:
        entries = await l_make.n_entries_list(loaded_feed_id, 3)

        another_entries_to_save = [entry.replace(feed_id=another_loaded_feed_id, id=uuid.uuid4()) for entry in entries]
        await l_domain.catalog_entries(reversed(another_entries_to_save))

        another_entries = await l_domain.get_entries_by_filter(feeds_ids=[another_loaded_feed_id], limit=100)

        user_a, user_b, user_c = await u_make.n_users(3)

        await m_domains.set_marker(user_a, Marker.read, another_entries[1].id)
        await m_domains.set_marker(user_b, Marker.read, entries[0].id)
        await m_domains.set_marker(user_b, Marker.read, another_entries[0].id)
        await m_domains.set_marker(user_b, Marker.read, entries[2].id)
        await m_domains.set_marker(user_c, Marker.read, entries[1].id)

        await merge_feeds(another_loaded_feed_id, loaded_feed_id)

        markers_a = await m_domains.get_markers(user_a, [entry.id for entry in another_entries])
        markers_b = await m_domains.get_markers(user_b, [entry.id for entry in another_entries])
        markers_c = await m_domains.get_markers(user_c, [entry.id for entry in another_entries])

        assert markers_a == {another_entries[1].id: {Marker.read}}
        assert markers_b == {another_entries[0].id: {Marker.read}, another_entries[2].id: {Marker.read}}
        assert markers_c == {another_entries[1].id: {Marker.read}}


class TestRemoveEntries:
    @pytest.mark.asyncio
    async def test_no_entries(self) -> None:
        await remove_entries([])

    @pytest.mark.asyncio
    async def test_success(
        self,
        loaded_feed_id: uuid.UUID,
        another_loaded_feed_id: uuid.UUID,
        fake_processor_id: int,
        another_fake_processor_id: int,
        three_processor_tags: tuple[ProcessorTag, ProcessorTag, ProcessorTag],
    ) -> None:
        entries = await l_make.n_entries_list(loaded_feed_id, 3)
        another_entries = await l_make.n_entries_list(another_loaded_feed_id, 3)

        tag_a, tag_b, tag_c = three_processor_tags

        # fill feed 1
        await o_domain.apply_tags_to_entry(entry_id=entries[0].id, processor_id=fake_processor_id, tags=[tag_a])

        await o_domain.apply_tags_to_entry(entry_id=entries[1].id, processor_id=fake_processor_id, tags=[tag_a])

        await o_domain.apply_tags_to_entry(
            entry_id=entries[1].id, processor_id=another_fake_processor_id, tags=[tag_b]
        )

        # fill feed 2
        await o_domain.apply_tags_to_entry(
            entry_id=another_entries[1].id, processor_id=fake_processor_id, tags=[tag_a]
        )

        await o_domain.apply_tags_to_entry(
            entry_id=another_entries[1].id, processor_id=another_fake_processor_id, tags=[tag_b]
        )

        await o_domain.apply_tags_to_entry(
            entry_id=another_entries[2].id, processor_id=another_fake_processor_id, tags=[tag_c]
        )

        await remove_entries([entries[0].id, another_entries[1].id, entries[2].id])

        loaded_entries = await l_domain.get_entries_by_ids(
            [entry.id for entry in entries] + [entry.id for entry in another_entries]
        )

        assert loaded_entries == {
            entries[0].id: None,
            entries[1].id: entries[1],
            entries[2].id: None,
            another_entries[0].id: another_entries[0],
            another_entries[1].id: None,
            another_entries[2].id: another_entries[2],
        }


class TestLimitEntriesForFeed:
    @pytest.mark.asyncio
    async def test_no_feed(self) -> None:
        await limit_entries_for_feed(uuid.uuid4(), limit=10)

    @pytest.mark.asyncio
    async def test_no_entries(self, loaded_feed_id: uuid.UUID) -> None:
        await limit_entries_for_feed(loaded_feed_id, limit=10)

    @pytest.mark.asyncio
    async def test_not_exceed_limit(self, loaded_feed_id: uuid.UUID) -> None:
        entries = await l_make.n_entries_list(loaded_feed_id, 3)

        await limit_entries_for_feed(loaded_feed_id, limit=10)

        loaded_entries = await l_domain.get_entries_by_filter(feeds_ids=[loaded_feed_id], limit=100)

        assert loaded_entries == entries

    @pytest.mark.asyncio
    async def test_exceed_limit(self, loaded_feed_id: uuid.UUID) -> None:
        entries = await l_make.n_entries_list(loaded_feed_id, 10)

        await limit_entries_for_feed(loaded_feed_id, limit=5)

        loaded_entries = await l_domain.get_entries_by_filter(feeds_ids=[loaded_feed_id], limit=100)

        assert loaded_entries == entries[:5]



from collections import defaultdict
from typing import Sequence

from ffun.scores import entities, operations


def get_score_contributions(rules: Sequence[entities.BaseRule], tags: set[int]) -> tuple[int, dict[int, int]]:
    score = 0
    contributions: dict[int, int] = defaultdict(int)

    for rule in rules:
        if rule.tags <= tags:
            score += rule.score
            for tag in rule.tags:
                contributions[tag] += rule.score

    return score, contributions


def get_score_rules(rules: list[entities.Rule], tags: set[int]) -> list[entities.Rule]:
    score_rules = []

    for rule in rules:
        if rule.tags <= tags:
            score_rules.append(rule)

    return score_rules


create_or_update_rule = operations.create_or_update_rule
delete_rule = operations.delete_rule
get_rules = operations.get_rules
update_rule = operations.update_rule



import uuid
from typing import Any

from ffun.feeds.entities import Feed, FeedState
from ffun.feeds.operations import get_feeds, save_feed


def fake_url() -> str:
    return f"https://{uuid.uuid4().hex}.com"


def fake_title() -> str:
    return f"Feed Title: {uuid.uuid4().hex}"


def fake_description() -> str:
    return f"Feed Description: {uuid.uuid4().hex}"


def fake_feed(**kwargs: Any) -> Feed:
    return Feed(
        id=uuid.uuid4() if "id" not in kwargs else kwargs["id"],
        url=fake_url() if "url" not in kwargs else kwargs["url"],
        state=FeedState.not_loaded if "state" not in kwargs else kwargs["state"],
        last_error=kwargs.get("last_error"),
        load_attempted_at=kwargs.get("load_attempted_at"),
        loaded_at=kwargs.get("loaded_at"),
        title=fake_title() if "title" not in kwargs else kwargs["title"],
        description=fake_description() if "description" not in kwargs else kwargs["description"],
    )


async def n_feeds(n: int) -> list[Feed]:
    feed_objects = [fake_feed() for _ in range(n)]

    saved_ids = [await save_feed(feed) for feed in feed_objects]

    feeds = await get_feeds(saved_ids)

    return feeds



import asyncio
from typing import Any

from ffun.core import logging
from ffun.core.background_tasks import InfiniteTask
from ffun.librarian import domain, operations
from ffun.librarian.processors.base import Processor
from ffun.librarian.processors.domain import Processor as DomainProcessor
from ffun.librarian.processors.native_tags import Processor as NativeTagsProcessor
from ffun.librarian.processors.openai_chat_3_5 import Processor as OpenAIChat35Processor
from ffun.librarian.processors.openai_chat_3_5_functions import Processor as OpenAIChat35FunctionsProcessor
from ffun.librarian.processors.upper_case_title import Processor as UpperCaseTitleProcessor
from ffun.librarian.settings import settings
from ffun.library import domain as l_domain

logger = logging.get_module_logger()


class ProcessorInfo:
    __slots__ = ("_id", "_processor", "_concurrency")

    def __init__(self, id: int, processor: Processor, concurrency: int):
        self._id = id
        self._processor = processor
        self._concurrency = concurrency

    @property
    def id(self) -> int:
        return self._id

    @property
    def processor(self) -> Processor:
        return self._processor

    @property
    def concurrency(self) -> int:
        return self._concurrency


processors = []

if settings.domain_processor.enabled:
    processors.append(
        ProcessorInfo(id=1, processor=DomainProcessor(name="domain"), concurrency=settings.domain_processor.workers)
    )


if settings.native_tags_processor.enabled:
    processors.append(
        ProcessorInfo(
            id=2, processor=NativeTagsProcessor(name="native_tags"), concurrency=settings.native_tags_processor.workers
        )
    )


if settings.openai_chat_35_processor.enabled:
    processors.append(
        ProcessorInfo(
            id=3,
            processor=OpenAIChat35Processor(name="openai_chat_3_5", model=settings.openai_chat_35_processor.model),
            concurrency=settings.openai_chat_35_processor.workers,
        )
    )


if settings.openai_chat_35_functions_processor.enabled:
    processors.append(
        ProcessorInfo(
            id=4,
            processor=OpenAIChat35FunctionsProcessor(
                name="openai_chat_3_5_functions", model=settings.openai_chat_35_functions_processor.model
            ),
            concurrency=settings.openai_chat_35_functions_processor.workers,
        )
    )

if settings.upper_case_title_processor.enabled:
    processors.append(
        ProcessorInfo(
            id=5,
            processor=UpperCaseTitleProcessor(name="upper_case_title"),
            concurrency=settings.upper_case_title_processor.workers,
        )
    )


class EntriesProcessor(InfiniteTask):
    __slots__ = ("_processor_info",)

    def __init__(self, processor_info: ProcessorInfo, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._processor_info = processor_info

    @property
    def id(self) -> int:
        return self._processor_info.id

    @property
    def concurrency(self) -> int:
        return self._processor_info.concurrency

    async def single_run(self) -> None:
        processor_id = self._processor_info.id
        concurrency = self._processor_info.concurrency

        # most likely, this call should be in a separate worker with a more complex logic
        # but for now it is ok to place it here
        await domain.plan_processor_queue(
            processor_id=processor_id,
            fill_when_below=concurrency,
            # TODO: move to settings
            chunk=concurrency * 10,
        )

        entities_ids = await operations.get_entries_to_process(processor_id=processor_id, limit=concurrency)

        if not entities_ids:
            logger.info("no_entries_to_process", processor_id=processor_id)
            return

        # TODO: we could add caching here, because multiple processors can request the same entries
        #       and currently we run multiple processors in the same process
        #       but remember that for API worker such cachin may be useless or even harmful
        entries = await l_domain.get_entries_by_ids(entities_ids)

        tasks = []
        entries_to_remove = []

        for entry_id, entry in entries.items():
            if entry is None:
                entries_to_remove.append(entry_id)
                continue

            tasks.append(
                domain.process_entry(processor_id=processor_id, processor=self._processor_info.processor, entry=entry)
            )

        if entries_to_remove:
            logger.warning("unexisted_entries_in_queue", processor_id=processor_id, entries_ids=entries_to_remove)
            tasks.append(
                domain.remove_entries_from_processor_queue(processor_id=processor_id, entry_ids=entries_to_remove)
            )

        await asyncio.gather(*tasks, return_exceptions=True)


def create_background_processors() -> list[EntriesProcessor]:
    return [
        EntriesProcessor(
            processor_info=processor_info,
            name=f"entries_processor_{processor_info.processor.name}",
            delay_between_runs=1,
        )
        for processor_info in processors
    ]



import json
import math
from typing import Any

import async_lru
import openai
import tiktoken
import typer

from ffun.core import logging
from ffun.openai import entities, errors
from ffun.openai.keys_statuses import statuses, track_key_status

logger = logging.get_module_logger()

cli = typer.Typer()


@async_lru.alru_cache()
async def get_encoding(model: str) -> tiktoken.Encoding:
    return tiktoken.encoding_for_model(model)


async def prepare_requests(
    system: str,  # pylint: disable=R0914
    text: str,
    model: str,
    function: dict[str, Any] | None,
    total_tokens: int,
    max_return_tokens: int,
) -> list[list[dict[str, str]]]:
    # high estimation on base of
    # https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb
    logger.info("prepare_requests")

    additional_tokens_per_message = 10

    encoding = await get_encoding(model)

    system_tokens = additional_tokens_per_message + len(encoding.encode("system")) + len(encoding.encode(system))

    text_tokens = additional_tokens_per_message + len(encoding.encode("user")) + len(encoding.encode(text))

    # rough estimation, because we don't know how exactly openai counts tokens for functions
    if function:
        function_tokens = (
            additional_tokens_per_message
            + len(encoding.encode("function"))
            + len(encoding.encode(json.dumps(function)))
        )
    else:
        function_tokens = 0

    tokens_per_chunk = (
        total_tokens - system_tokens - max_return_tokens - function_tokens - additional_tokens_per_message
    )

    if text_tokens <= tokens_per_chunk:
        logger.info("single_chunk_text")
        return [[{"role": "system", "content": system}, {"role": "user", "content": text}]]

    logger.info("multiple_chunks_text", text_tokens=text_tokens, tokens_per_chunk=tokens_per_chunk)

    messages = []

    # TODO: should we split text on chunks with intersections?
    # TODO: what to do with the last small chunks?
    expected_chunks_number = text_tokens // tokens_per_chunk + 1

    expected_chunk_size = int(math.floor(len(text) / expected_chunks_number))

    logger.info(
        "expected_chunks", expected_chunks_number=expected_chunks_number, expected_chunk_size=expected_chunk_size
    )

    offset = 0
    current_chunk_size = expected_chunk_size

    # expected_chunk_size is not accurate,
    # because tokens have different length in charactets
    # => number of characters to send is depends on type of part of the text,
    #    where they were taken from
    while offset < len(text):
        expected_chunk = text[offset : offset + current_chunk_size]

        expected_text_tokens = additional_tokens_per_message + len(encoding.encode(expected_chunk))

        if expected_text_tokens > tokens_per_chunk:
            current_chunk_size = math.ceil(current_chunk_size * 0.95)
            continue

        messages.append([{"role": "system", "content": system}, {"role": "user", "content": expected_chunk}])

        offset += current_chunk_size
        current_chunk_size = expected_chunk_size

    return messages


async def request(  # noqa: CFQ002
    api_key: str,
    model: str,
    messages: list[dict[str, str]],
    function: dict[str, Any] | None,
    max_tokens: int,
    temperature: float,
    top_p: float,
    presence_penalty: float,
    frequency_penalty: float,
) -> entities.OpenAIAnswer:
    logger.info("request_openai")

    arguments: dict[str, Any] = {}

    if function is not None:
        arguments["functions"] = [function]
        arguments["function_call"] = {"name": function["name"]}

    with track_key_status(api_key):
        try:
            answer = await openai.ChatCompletion.acreate(
                api_key=api_key,
                model=model,
                temperature=temperature,
                max_tokens=max_tokens,
                top_p=top_p,
                presence_penalty=presence_penalty,
                frequency_penalty=frequency_penalty,
                messages=messages,
                **arguments
            )
        except openai.error.APIError as e:
            logger.info("openai_api_error", message=str(e))
            raise errors.TemporaryError(message=str(e)) from e

    logger.info("openai_response")

    if function:
        content = answer["choices"][0]["message"]["function_call"]["arguments"]
    else:
        content = answer["choices"][0]["message"]["content"]

    return entities.OpenAIAnswer(
        content=content,
        prompt_tokens=answer["usage"]["prompt_tokens"],
        completion_tokens=answer["usage"]["completion_tokens"],
        total_tokens=answer["usage"]["total_tokens"],
    )


async def multiple_requests(  # noqa: CFQ002
    api_key: str,
    model: str,
    messages: list[list[dict[str, str]]],
    function: dict[str, Any] | None,
    max_return_tokens: int,
    temperature: float,
    top_p: float,
    presence_penalty: float,
    frequency_penalty: float,
) -> list[entities.OpenAIAnswer]:
    # TODO: rewrite to gather
    #       also, it seems OpenAI API support sending multiple threads in a single request
    results = []

    for i, request_messages in enumerate(messages):
        logger.info("request", number=i, total=len(messages))
        result = await request(
            api_key=api_key,
            model=model,
            messages=request_messages,
            function=function,
            max_tokens=max_return_tokens,
            temperature=temperature,
            top_p=top_p,
            presence_penalty=presence_penalty,
            frequency_penalty=frequency_penalty,
        )

        results.append(result)

    return results


async def check_api_key(api_key: str) -> entities.KeyStatus:
    with track_key_status(api_key):
        try:
            await openai.Model.alist(api_key=api_key)
        except openai.error.APIError:
            pass

    return statuses.get(api_key)



import pytest

from ffun.feeds.entities import Feed
from ffun.library import operations
from ffun.library.domain import get_entry, normalize_entry
from ffun.library.entities import Entry, EntryChange


class TestNormalizeEntry:
    @pytest.mark.asyncio
    async def test_no_changes(self, cataloged_entry: Entry) -> None:
        changes = await normalize_entry(cataloged_entry)
        assert changes == []

        loaded_entry = await get_entry(cataloged_entry.id)
        assert loaded_entry == cataloged_entry

    @pytest.mark.parametrize("apply", [True, False])
    @pytest.mark.asyncio
    async def test_normalize_external_url(self, new_entry: Entry, loaded_feed: Feed, apply: bool) -> None:
        wrong_url = "/relative/url"
        expected_url = f"{loaded_feed.url}{wrong_url}"

        new_entry = new_entry.replace(external_url=wrong_url)

        await operations.catalog_entries([new_entry])

        entry = await get_entry(new_entry.id)

        changes = await normalize_entry(entry, apply=apply)

        assert changes == [
            EntryChange(
                id=new_entry.id, field="external_url", old_value=new_entry.external_url, new_value=expected_url
            )
        ]

        loaded_entry = await get_entry(new_entry.id)

        if apply:
            assert loaded_entry.external_url == expected_url
        else:
            assert loaded_entry.external_url == wrong_url



import functools
import random
import uuid

import pytest
import pytest_asyncio

from ffun.feeds.domain import save_feed
from ffun.feeds.entities import Feed
from ffun.feeds.tests import make as f_make
from ffun.feeds_collections import predefines


@functools.cache
def all_urls() -> set[str]:
    urls = set()

    for collection in predefines.predefines.values():
        urls.update(collection)

    return urls


@pytest.fixture
def new_collection_feed() -> Feed:
    return f_make.fake_feed(url=random.choice(list(all_urls())))


@pytest_asyncio.fixture
async def saved_collection_feed_id(new_collection_feed: Feed) -> uuid.UUID:
    return await save_feed(new_collection_feed)



import uuid
from typing import Any

from ffun.core import utils
from ffun.parsers.entities import EntryInfo


def fake_url() -> str:
    return f"https://{uuid.uuid4().hex}.com"


def fake_title() -> str:
    return f"Entry Title: {uuid.uuid4().hex}"


def fake_body() -> str:
    return f"Entry Body: {uuid.uuid4().hex}"


def fake_entry_info(**kwargs: Any) -> EntryInfo:
    return EntryInfo(
        title=fake_title() if "title" not in kwargs else kwargs["title"],
        body=fake_body() if "body" not in kwargs else kwargs["body"],
        external_id=uuid.uuid4().hex if "external_id" not in kwargs else kwargs["external_id"],
        external_url=fake_url() if "external_url" not in kwargs else kwargs["external_url"],
        external_tags={uuid.uuid4().hex, uuid.uuid4().hex}
        if "external_tags" not in kwargs
        else kwargs["external_tags"],
        published_at=utils.now() if "published_at" not in kwargs else kwargs["published_at"],
    )



from typing import Any

import fastapi
from fastapi.responses import JSONResponse
from sentry_sdk import capture_exception

from ffun.core import api, logging
from ffun.core.errors import Error

logger = logging.get_module_logger()


async def _handle_expected_error(_: fastapi.Request, error: Error) -> JSONResponse:
    # TODO: improve error processing
    api_error = api.APIError(code=error.__class__.__name__)

    capture_exception(error)

    logger.exception(error.__class__.__name__, sentry_skip=True)

    return JSONResponse(status_code=500, content=api_error.dict())


async def _handle_unexpected_error(_: fastapi.Request, error: Exception) -> JSONResponse:
    # TODO: improve error processing
    api_error = api.APIError(
        code=error.__class__.__name__, message="An unexpected error appeared. We are working on fixing it."
    )

    capture_exception(error)

    logger.exception(error.__class__.__name__, sentry_skip=True)

    return JSONResponse(status_code=500, content=api_error.dict())


# TODO: move somewhere?
def initialize_error_processors(app: fastapi.FastAPI) -> fastapi.FastAPI:
    app.exception_handler(Error)(_handle_expected_error)
    return app


async def final_errors_middleware(request: fastapi.Request, call_next: Any) -> fastapi.Response:
    try:
        return await call_next(request)  # type: ignore
    except Error as e:
        return await _handle_expected_error(request, e)
    except Exception as e:
        return await _handle_unexpected_error(request, e)



import contextlib
import datetime
import uuid
from typing import Any, AsyncGenerator, Iterable

from ffun.core import logging
from ffun.feeds_collections import domain as fc_domain
from ffun.feeds_links import domain as fl_domain
from ffun.openai import client, entities, errors
from ffun.openai.keys_statuses import statuses
from ffun.resources import domain as r_domain
from ffun.user_settings import domain as us_domain

logger = logging.get_module_logger()


# Note: this code is not about billing, it is about protection from the overuse of keys
#       i.e. in case of a problem, we should count a key as used with a maximum used tokens


# TODO: add lock here to not check the same key in parallel by different processors
async def _api_key_is_working(api_key: str) -> bool:
    status = statuses.get(api_key)

    if status == entities.KeyStatus.works:
        return True

    if status != entities.KeyStatus.unknown:
        return False

    new_status = await client.check_api_key(api_key)

    return new_status == entities.KeyStatus.works


async def _filter_out_users_with_wrong_keys(
    infos: list[entities.UserKeyInfo], **kwargs: Any
) -> list[entities.UserKeyInfo]:
    return [info for info in infos if info.api_key and await _api_key_is_working(info.api_key)]


async def _filter_out_users_without_keys(
    infos: list[entities.UserKeyInfo], **kwargs: Any
) -> list[entities.UserKeyInfo]:
    return [info for info in infos if info.api_key]


async def _filter_out_users_for_whome_entry_is_too_old(
    infos: list[entities.UserKeyInfo], entry_age: datetime.timedelta, **kwargs: Any
) -> list[entities.UserKeyInfo]:
    return [info for info in infos if info.process_entries_not_older_than >= entry_age]


async def _filter_out_users_with_overused_keys(
    infos: list[entities.UserKeyInfo], reserved_tokens: int, **kwargs: Any
) -> list[entities.UserKeyInfo]:
    return [info for info in infos if info.tokens_used + reserved_tokens < info.max_tokens_in_month]


async def _users_for_feed(feed_id: uuid.UUID) -> set[uuid.UUID]:
    from ffun.application.user_settings import UserSetting

    user_ids = await fl_domain.get_linked_users(feed_id)

    # find all users for collections
    if await fc_domain.is_feed_in_collections(feed_id):
        collections_user_ids = await us_domain.get_users_with_setting(
            UserSetting.openai_allow_use_key_for_collections, True
        )
        user_ids.extend(collections_user_ids)

    return set(user_ids)


async def _choose_user(
    infos: list[entities.UserKeyInfo], reserved_tokens: int, interval_started_at: datetime.datetime
) -> entities.UserKeyInfo:
    from ffun.application.resources import Resource as AppResource

    for info in infos:
        if await r_domain.try_to_reserve(
            user_id=info.user_id,
            kind=AppResource.openai_tokens,
            interval_started_at=interval_started_at,
            amount=reserved_tokens,
            limit=info.max_tokens_in_month,
        ):
            return info

    raise errors.NoKeyFoundForFeed()


@contextlib.asynccontextmanager
async def _use_key(
    user_id: uuid.UUID, api_key: str, reserved_tokens: int, interval_started_at: datetime.datetime
) -> AsyncGenerator[entities.APIKeyUsage, None]:
    from ffun.application.resources import Resource as AppResource

    log = logger.bind(function="_use_key", user_id=user_id)

    key_usage = entities.APIKeyUsage(user_id=user_id, api_key=api_key, used_tokens=None)

    used_tokens = reserved_tokens

    try:
        log.info("provide_key")

        yield key_usage

        assert key_usage.used_tokens is not None

        used_tokens = key_usage.used_tokens

        log.info("key_used", used_tokens=used_tokens)

    finally:
        log.info("convert_reserved_to_used", reserved_tokens=reserved_tokens, used_tokens=used_tokens)

        await r_domain.convert_reserved_to_used(
            user_id=user_id,
            kind=AppResource.openai_tokens,
            interval_started_at=interval_started_at,
            used=used_tokens,
            reserved=reserved_tokens,
        )

        log.info("resources_converted")


async def _get_user_key_infos(
    user_ids: Iterable[uuid.UUID], interval_started_at: datetime.datetime
) -> list[entities.UserKeyInfo]:
    from ffun.application.resources import Resource as AppResource
    from ffun.application.user_settings import UserSetting

    users_settings = await us_domain.load_settings_for_users(
        user_ids,
        kinds=[
            UserSetting.openai_api_key,
            UserSetting.openai_max_tokens_in_month,
            UserSetting.openai_process_entries_not_older_than,
        ],
    )

    resources = await r_domain.load_resources(
        user_ids=user_ids, kind=AppResource.openai_tokens, interval_started_at=interval_started_at
    )

    infos = []

    for user_id in user_ids:
        settings = users_settings[user_id]

        days = settings.get(UserSetting.openai_process_entries_not_older_than)

        assert days is not None
        assert isinstance(days, int)

        max_tokens_in_month = settings.get(UserSetting.openai_max_tokens_in_month)

        assert max_tokens_in_month is not None
        assert isinstance(max_tokens_in_month, int)

        infos.append(
            entities.UserKeyInfo(
                user_id=user_id,
                api_key=settings.get(UserSetting.openai_api_key),
                max_tokens_in_month=max_tokens_in_month,
                process_entries_not_older_than=datetime.timedelta(days=days),
                tokens_used=resources[user_id].total,
            )
        )

    return infos


_filters = (
    _filter_out_users_without_keys,
    _filter_out_users_for_whome_entry_is_too_old,
    _filter_out_users_with_wrong_keys,
    _filter_out_users_with_overused_keys,
)


async def _get_candidates(
    feed_id: uuid.UUID,
    interval_started_at: datetime.datetime,
    entry_age: datetime.timedelta,
    reserved_tokens: int,
    filters: tuple[Any, ...] = _filters,
) -> list[entities.UserKeyInfo]:
    user_ids = list(await _users_for_feed(feed_id))

    infos = await _get_user_key_infos(user_ids, interval_started_at)

    for _filter in filters:
        if not infos:
            return []

        infos = await _filter(infos=infos, entry_age=entry_age, reserved_tokens=reserved_tokens)

    return infos


async def _find_best_user_with_key(
    feed_id: uuid.UUID, entry_age: datetime.timedelta, interval_started_at: datetime.datetime, reserved_tokens: int
) -> entities.UserKeyInfo:
    infos = await _get_candidates(
        feed_id=feed_id, interval_started_at=interval_started_at, entry_age=entry_age, reserved_tokens=reserved_tokens
    )

    infos.sort(key=lambda info: info.tokens_used)

    return await _choose_user(infos=infos, reserved_tokens=reserved_tokens, interval_started_at=interval_started_at)


@contextlib.asynccontextmanager
async def api_key_for_feed_entry(  # noqa: CCR001,CFQ001
    feed_id: uuid.UUID, entry_age: datetime.timedelta, reserved_tokens: int
) -> AsyncGenerator[entities.APIKeyUsage, None]:
    interval_started_at = r_domain.month_interval_start()

    found_user = await _find_best_user_with_key(
        feed_id=feed_id, entry_age=entry_age, interval_started_at=interval_started_at, reserved_tokens=reserved_tokens
    )

    assert found_user.api_key is not None

    async with _use_key(
        user_id=found_user.user_id,
        api_key=found_user.api_key,
        reserved_tokens=reserved_tokens,
        interval_started_at=interval_started_at,
    ) as key_usage:
        yield key_usage



import datetime
import uuid

import pytest

from ffun.resources.domain import load_resource, month_interval_start
from ffun.resources.operations import initialize_resource, try_to_reserve

_kind = 214
_another_kind = 215


class TestMonthIntervalStart:
    @pytest.mark.asyncio
    async def test_month_interval_start(self) -> None:
        now = datetime.datetime.now()

        assert month_interval_start(now) == datetime.datetime(
            year=now.year,
            month=now.month,
            day=1,
            hour=0,
            minute=0,
            second=0,
            microsecond=0,
            tzinfo=now.tzinfo,
        )


class TestLoadResource:
    @pytest.mark.asyncio
    async def test_initialized(self, internal_user_id: uuid.UUID) -> None:
        interval_started_at = month_interval_start()

        await initialize_resource(user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at)

        await try_to_reserve(
            user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at, amount=13, limit=100
        )

        resource = await load_resource(user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at)

        assert resource.user_id == internal_user_id
        assert resource.kind == _kind
        assert resource.interval_started_at == interval_started_at
        assert resource.used == 0
        assert resource.reserved == 13

    @pytest.mark.asyncio
    async def test_not_initialized(self, internal_user_id: uuid.UUID) -> None:
        interval_started_at = month_interval_start()

        resource = await load_resource(user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at)

        assert resource.user_id == internal_user_id
        assert resource.kind == _kind
        assert resource.interval_started_at == interval_started_at
        assert resource.used == 0
        assert resource.reserved == 0



import datetime
from typing import Type

import openai
import pytest
from pytest_mock import MockerFixture

from ffun.openai.entities import KeyStatus
from ffun.openai.keys_statuses import Statuses, track_key_status


class TestStatuses:
    def test_inititalization(self, statuses: Statuses) -> None:
        assert statuses._statuses == {}

    def test_set_status(self, statuses: Statuses) -> None:
        status_1, status_2, *_ = list(KeyStatus)

        statuses.set("key_1", status_1)
        statuses.set("key_2", status_2)

        assert statuses.get("key_1") == status_1
        assert statuses.get("key_2") == status_2

    def test_get__no_key(self, statuses: Statuses) -> None:
        assert statuses.get("key_1") == KeyStatus.unknown

    def test_get__works(self, statuses: Statuses) -> None:
        statuses.set("key_1", KeyStatus.works)
        assert statuses.get("key_1") == KeyStatus.works

    def test_get__broken(self, statuses: Statuses) -> None:
        statuses.set("key_1", KeyStatus.broken)
        assert statuses.get("key_1") == KeyStatus.broken

    def test_get__broken_expired(self, statuses: Statuses, mocker: MockerFixture) -> None:
        statuses.set("key_1", KeyStatus.broken)

        mocker.patch("ffun.openai.settings.settings.key_broken_timeout", datetime.timedelta(seconds=0))

        assert statuses.get("key_1") == KeyStatus.unknown

    def test_get__quota(self, statuses: Statuses) -> None:
        statuses.set("key_1", KeyStatus.quota)
        assert statuses.get("key_1") == KeyStatus.quota

    def test_get__quota_expired(self, statuses: Statuses, mocker: MockerFixture) -> None:
        statuses.set("key_1", KeyStatus.quota)

        mocker.patch("ffun.openai.settings.settings.key_quota_timeout", datetime.timedelta(seconds=0))

        assert statuses.get("key_1") == KeyStatus.unknown


class TestTrackKeyStatus:
    def test_works(self, statuses: Statuses) -> None:
        with track_key_status("key_1", statuses):
            pass

        assert statuses.get("key_1") == KeyStatus.works

    @pytest.mark.parametrize(
        "exception", [openai.error.AuthenticationError, openai.error.PermissionError, openai.error.InvalidAPIType]
    )
    def test_authentication_error(self, exception: Type[Exception], statuses: Statuses) -> None:
        with pytest.raises(exception):
            with track_key_status("key_1", statuses):
                raise exception()

        assert statuses.get("key_1") == KeyStatus.broken

    @pytest.mark.parametrize("exception", [openai.error.RateLimitError])
    def test_quota_error(self, exception: Type[Exception], statuses: Statuses) -> None:
        with pytest.raises(exception):
            with track_key_status("key_1", statuses):
                raise exception()

        assert statuses.get("key_1") == KeyStatus.quota



import uuid

import pytest
import pytest_asyncio

from ffun.ontology.domain import get_ids_by_uids
from ffun.ontology.entities import ProcessorTag


@pytest_asyncio.fixture
async def three_tags_by_uids() -> dict[str, int]:
    return await get_ids_by_uids([uuid.uuid4().hex for _ in range(3)])


@pytest.fixture
def three_tags_by_ids(three_tags_by_uids: dict[str, int]) -> dict[int, str]:
    return {v: k for k, v in three_tags_by_uids.items()}


@pytest.fixture
def three_tags_ids(three_tags_by_uids: dict[str, int]) -> tuple[int, int, int]:
    return tuple(three_tags_by_uids.values())  # type: ignore


@pytest.fixture
def three_processor_tags(three_tags_by_ids: dict[int, str]) -> tuple[ProcessorTag, ProcessorTag, ProcessorTag]:
    return tuple(ProcessorTag(raw_uid=tag) for tag in three_tags_by_ids.values())  # type: ignore
